import JSON5 from 'json5';
import { ConfiguredVueProject, InferredVueProject, createFullAnalyzer, transformToFunction } from '@vuedx/analyze';
import { VueTextDocument, isVirtualFile, getContainingFile, DocumentStore, asUri, isVueFile, MODULE_SELECTOR, parseVirtualFileName, VIRTUAL_FILENAME_SEPARATOR, INTERNAL_MODULE_SELECTOR, RENDER_SELECTOR, SCRIPT_SETUP_BLOCK_SELECTOR, SCRIPT_BLOCK_SELECTOR, isVirtualFileOfType } from '@vuedx/vue-virtual-textdocument';
import Path from 'path';
import { URI } from 'vscode-uri';
import { inspect } from 'util';
import { findTemplateNodeAt, findTemplateNodeInRange, findTemplateChildrenInRange, isSimpleIdentifier, isElementNode, isDirectiveNode, traverseFast, isComponentNode, isSimpleExpressionNode, createSimpleExpression, stringify, isAttributeNode, isTextNode, isPlainElementNode, traverseEvery } from '@vuedx/template-ast-types';
import QuickLRU from 'quick-lru';

function wrapInTrace(context, target) {
    return target;
}
function wrapFn(context, target) {
    return target;
}

const PATCHED_METHODS = Symbol('Vue Patched Methods');
function tryPatchMethod(target, methodName, createOverride) {
    const patched = target[PATCHED_METHODS] || [];
    if (patched.includes(methodName))
        return;
    let fn = target[methodName];
    if (typeof fn === 'function') {
        try {
            fn = fn.bind(target);
        }
        catch (_a) {
            // - ignore
        }
    }
    target[methodName] = createOverride(fn);
    patched.push(methodName);
    target[PATCHED_METHODS] = patched;
}

function getLastNumberFromVersion(version) {
    const parts = version.split(/[^0-9]+/);
    const ver = parts.pop();
    return Number(ver);
}
class ProxyDocumentStore extends DocumentStore {
    get(fileNameOrUri) {
        return super.get(asUri(fileNameOrUri));
    }
    has(fileNameOrUri) {
        return super.has(asUri(fileNameOrUri));
    }
    set(fileNameOrUri, document) {
        return super.set(asUri(fileNameOrUri), document);
    }
    delete(fileNameOrUri) {
        return super.delete(asUri(fileNameOrUri));
    }
}
function getConfig(config = {}) {
    return {
        ...config,
        features: {
            diagnostics: true,
            organizeImports: true,
            quickInfo: true,
            rename: true,
            refactor: true,
            goto: true,
            tagCompletions: true,
            ...config.features,
        },
    };
}
class PluginContext {
    constructor(typescript) {
        this.typescript = typescript;
        this._config = getConfig();
        this._externalFiles = new WeakMap();
        this._vueProjects = [];
        this.lastCleanupAt = Date.now();
        this.store = new ProxyDocumentStore((uri) => {
            var _a;
            const fileName = URI.parse(uri).fsPath;
            const content = (_a = this.typescript.sys.readFile(fileName)) !== null && _a !== void 0 ? _a : '';
            return this.createVueDocument(fileName, content);
        }, (uri) => {
            const fileName = URI.parse(uri).fsPath;
            return URI.file(this.projectService.toPath(fileName)).toString();
        });
    }
    get config() {
        return this._config;
    }
    get serviceHost() {
        return this._serverHost;
    }
    get projectService() {
        return this._projectService;
    }
    log(message) {
        if (this.projectService != null) {
            this.projectService.logger.info(`Vue.js:: ${message}`);
        }
    }
    getVueVersion(fileName) {
        var _a, _b;
        return (_b = (_a = this.getVueProjectForFile(fileName, false)) === null || _a === void 0 ? void 0 : _a.version) !== null && _b !== void 0 ? _b : '3.0.0';
    }
    getExternalFiles(project) {
        var _a;
        return (_a = this._externalFiles.get(project)) !== null && _a !== void 0 ? _a : [];
    }
    forEachTSProject(fn) {
        this.projectService.configuredProjects.forEach(fn);
        this.projectService.inferredProjects.forEach(fn);
        this.projectService.externalProjects.forEach(fn);
    }
    getVueProjectForFile(fileName, ensure) {
        var _a, _b, _c;
        let project = (_b = (_a = this._vueProjects.find(({ project }) => fileName.startsWith(project.rootDir + '/'))) === null || _a === void 0 ? void 0 : _a.project) !== null && _b !== void 0 ? _b : null;
        if (project === null && ensure === true) {
            const packageFile = this.typescript.findConfigFile(fileName, this.typescript.sys.fileExists, 'package.json');
            const configFile = this.typescript.findConfigFile(fileName, this.typescript.sys.fileExists, 'vueconfig.json');
            const rootDir = Path.posix.dirname((_c = packageFile !== null && packageFile !== void 0 ? packageFile : configFile) !== null && _c !== void 0 ? _c : fileName);
            const fileNames = this.serviceHost.readDirectory(rootDir, ['.tsx', '.jsx', '.ts', '.js'], ['node_modules']);
            const tryRequire = (fileName) => {
                var _a;
                try {
                    if (Path.posix.basename(fileName) === 'vueconfig.json') {
                        const contents = (_a = this.serviceHost.readFile(fileName)) !== null && _a !== void 0 ? _a : '';
                        return JSON5.parse(contents);
                    }
                    // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
                    delete require.cache[fileName];
                    return require(fileName);
                }
                catch (error) {
                    return {};
                }
            };
            const newProject = configFile != null
                ? new ConfiguredVueProject(rootDir, packageFile, packageFile != null ? tryRequire(packageFile) : {}, configFile, tryRequire(configFile))
                : new InferredVueProject(rootDir, packageFile, packageFile != null ? tryRequire(packageFile) : {});
            newProject.setFileNames(fileNames);
            const disposables = [
                () => {
                    this._vueProjects.splice(this._vueProjects.findIndex((item) => item.project === newProject), 1);
                },
            ];
            const dispose = () => disposables.forEach((fn) => fn());
            const projectDirWatcher = this.serviceHost.watchDirectory(rootDir, (fileName) => {
                if (fileName.endsWith('/vueconfig.json')) {
                    if (newProject.kind === 'inferred') {
                        dispose();
                        reload();
                        return;
                    }
                }
                if (!/\.(vue|ts|tsx|js|jsx)$/.test(fileName))
                    return;
                if (this.serviceHost.fileExists(fileName)) {
                    if (!newProject.fileNames.includes(fileName)) {
                        newProject.setFileNames([...newProject.fileNames, fileName]);
                    }
                }
                else {
                    if (newProject.fileNames.includes(fileName)) {
                        newProject.setFileNames(newProject.fileNames.filter((f) => f !== fileName));
                    }
                }
            });
            disposables.push(() => projectDirWatcher.close());
            const reload = () => {
                this.forEachTSProject((tsProject) => {
                    const dir = tsProject.getCurrentDirectory();
                    if (dir.startsWith(newProject.rootDir) ||
                        newProject.rootDir.startsWith(dir)) {
                        this.getExternalFiles(tsProject).forEach((fileName) => {
                            const document = this.store.get(fileName);
                            if (document != null) {
                                document.markDirty();
                                document.all().forEach((doc) => {
                                    if ([
                                        'javascript',
                                        'typescript',
                                        'javascriptreact',
                                        'typescriptreact',
                                    ].includes(doc.languageId)) {
                                        triggerFileUpdate(this, doc.fsPath);
                                    }
                                });
                            }
                        });
                        tsProject.markAsDirty();
                        tsProject.refreshDiagnostics();
                    }
                });
            };
            if (configFile != null) {
                const configFileWatcher = this.serviceHost.watchFile(configFile, (_, event) => {
                    if (event === this.typescript.FileWatcherEventKind.Deleted) {
                        dispose();
                    }
                    else {
                        newProject.setConfig(tryRequire(configFile));
                        newProject.markDirty();
                    }
                    reload();
                });
                disposables.push(() => {
                    configFileWatcher.close();
                });
            }
            if (packageFile != null) {
                const packageFileWatcher = this.serviceHost.watchFile(packageFile, (_, event) => {
                    if (event === this.typescript.FileWatcherEventKind.Deleted) {
                        dispose();
                    }
                    else {
                        newProject.packageJSON = {
                            dependencies: {},
                            devDependencies: {},
                            ...tryRequire(packageFile),
                        };
                        newProject.markDirty();
                    }
                });
                disposables.push(() => packageFileWatcher.close());
            }
            this._vueProjects.push({ project: newProject, dispose });
            project = newProject;
        }
        return project;
    }
    disposeUnusedProjects(force = false) {
        if (!force &&
            (Date.now() - this.lastCleanupAt > 5 * 60 * 1000 ||
                this._vueProjects.length === 0)) {
            return;
        }
        this.lastCleanupAt = Date.now();
        const usedProjects = new Set();
        // TODO: detect unused projects
        this._vueProjects.forEach(({ project, dispose }) => {
            if (!usedProjects.has(project)) ;
        });
    }
    hasAnyVueReference(fileName) {
        // TODO: check if there is an import from .vue file or not.
        // This could be used in routing service to skip some work when .vue
        // is not in the play.
        return true;
    }
    createVueDocument(fileName, content) {
        const uri = URI.file(fileName).toString();
        const document = VueTextDocument.create(uri, 'vue', 0, content, {
            getGlobalComponents: wrapFn('getGlobalComponents', () => {
                const project = this.getVueProjectForFile(fileName, true);
                return project.kind === 'inferred'
                    ? project.components
                    : project.globalComponents;
            }),
        });
        this.store.set(uri, document);
        return document;
    }
    error(error) {
        var _a;
        if (this.projectService != null) {
            this.projectService.logger.msg(`Vue.js:: ${error.message} ${(_a = error.stack) !== null && _a !== void 0 ? _a : ''}`, this.typescript.server.Msg.Err);
        }
    }
    tryCreateScriptInfo(fileName) {
        try {
            if (isVirtualFile(fileName)) {
                const vueFileName = getContainingFile(fileName);
                this.store.get(vueFileName);
                const scriptInfo = this.projectService.getOrCreateScriptInfoForNormalizedPath(this.typescript.server.toNormalizedPath(vueFileName), false);
                if (scriptInfo != null) {
                    patchScriptInfo(this, scriptInfo);
                }
            }
        }
        catch (_a) { }
    }
    load(info) {
        this.log(`Loading Vue plugin: ${info.project.getProjectName()}`);
        this._serverHost = info.serverHost;
        this._projectService = info.project.projectService;
        this.setConfig(info.config);
        patchProjectService(this);
        patchServiceHost(this);
        patchLanguageServiceHost(this, info.languageServiceHost);
    }
    setConfig(config) {
        this._config = getConfig(config);
    }
}
function patchProjectService(context) {
    patchExtraFileExtensions(context);
}
function patchExtraFileExtensions(context) {
    var _a;
    const extraFileExtensions = [
        {
            extension: 'vue',
            isMixedContent: false,
            scriptKind: context.typescript.ScriptKind.Deferred,
        },
    ];
    tryPatchMethod(context.projectService, 'setHostConfiguration', (setHostConfiguration) => {
        context.log(`[patch] Add support for vue extension. (ProjectService)`);
        return wrapFn('setHostConfiguration', (args) => {
            const current = context.projectService
                .hostConfiguration
                .extraFileExtensions;
            if (args.extraFileExtensions != null) {
                args.extraFileExtensions.push(...extraFileExtensions);
                context.log(`extraFileExtensions: ${JSON.stringify(args.extraFileExtensions)}`);
            }
            else if (current == null ||
                !current.some((ext) => ext.extension === 'vue')) {
                args.extraFileExtensions = [...extraFileExtensions];
            }
            return setHostConfiguration(args);
        });
    });
    if (((_a = context.projectService
        .hostConfiguration.extraFileExtensions) === null || _a === void 0 ? void 0 : _a.some((ext) => ext.extension === 'vue')) === true) {
        return;
    }
    // Enable .vue after enhancing the language server.
    context.projectService.setHostConfiguration({ extraFileExtensions: [] });
}
function patchLanguageServiceHost(context, languageServiceHost) {
    patchGetScriptFileNames(context, languageServiceHost);
    patchModuleResolution(context, languageServiceHost);
    tryPatchMethod(languageServiceHost, 'getCompilationSettings', (getCompilationSettings) => () => {
        const settings = getCompilationSettings();
        settings.jsx = context.typescript.JsxEmit.Preserve;
        if (settings.allowJs !== false) {
            settings.allowJs = true;
        }
        if (settings.checkJs !== false) {
            settings.checkJs = true;
        }
        settings.noEmit = true;
        return settings;
    });
}
function patchGetScriptFileNames(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'getScriptFileNames', (getScriptFileNames) => {
        context.log(`[patch] Override getScriptFileNames to expand .vue files to virtual files. (LanguageServerHost)`);
        return wrapFn('getScriptFileNames', () => {
            const fileNames = new Set();
            const vueFiles = new Set();
            getScriptFileNames()
                .map((fileName) => isVirtualFile(fileName) ? getContainingFile(fileName) : fileName)
                .forEach((fileName) => {
                if (isVueFile(fileName)) {
                    const document = context.store.get(fileName);
                    if (document != null) {
                        vueFiles.add(fileName);
                        fileNames.add(document.getDocumentFileName(MODULE_SELECTOR));
                    }
                }
                else {
                    fileNames.add(fileName);
                }
            });
            if (vueFiles.size > 0) {
                const files = Array.from(vueFiles);
                const projects = new Set(files
                    .map((fileName) => context.projectService.getDefaultProjectForFile(context.typescript.server.toNormalizedPath(fileName), false))
                    .filter(Boolean));
                projects.forEach((project) => {
                    context._externalFiles.set(project, files);
                    if (project instanceof context.typescript.server.ConfiguredProject) {
                        const vueProject = context.getVueProjectForFile(project.getCurrentDirectory() + '/placeholder.vue', true);
                        vueProject.setFileNames(files);
                    }
                });
            }
            return Array.from(fileNames);
        });
    });
}
function patchFileExists(context) {
    tryPatchMethod(context.serviceHost, 'fileExists', (fileExists = context.typescript.sys.fileExists) => {
        context.log(`[patch] Override fileExists to check containing file for virtual files. (ServiceHost)`);
        return wrapFn('fileExists', (fileName) => {
            if (isVirtualFile(fileName)) {
                const document = context.store.get(getContainingFile(fileName));
                const result = parseVirtualFileName(fileName);
                return (document != null &&
                    result != null &&
                    fileName === document.getDocumentFileName(result.selector));
            }
            return fileExists(fileName);
        });
    });
}
function patchReadFile(context) {
    tryPatchMethod(context.serviceHost, 'readFile', (readFile) => {
        context.log(`[patch] Override readFile to check containing file for virtual files. (ServiceHost)`);
        return wrapFn('readFile', (fileName, encoding) => {
            var _a;
            if (isVirtualFile(fileName)) {
                context.log(`host.readFile("${fileName}")`);
                const document = (_a = context.store
                    .get(getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName);
                if (document != null)
                    return document.getText();
                else
                    return;
            }
            return readFile != null
                ? readFile(fileName, encoding)
                : context.typescript.sys.readFile(fileName, encoding);
        });
    });
}
function patchReadDirectory(context) {
    tryPatchMethod(context.serviceHost, 'readDirectory', (readDirectory) => {
        return wrapFn('readDirectory', (path, extensions, exclude, include, depth) => {
            if (extensions != null) {
                extensions = [...extensions, '.vue'];
            }
            return readDirectory(path, extensions, exclude, include, depth);
        });
    });
}
function patchModuleResolution(context, languageServiceHost) {
    tryPatchMethod(languageServiceHost, 'resolveModuleNames', (resolveModuleNames) => {
        context.log(`[patch] Override resolveModuleNames to resolve imports from .vue files. (LanguageServerHost)`);
        return wrapFn('resolveModuleNames', (moduleNames, containingFile, reusedNames, redirectedReferences, options) => {
            if (isVueFile(containingFile))
                throw new Error('A .vue file should not be part of TS program.');
            const newModuleNames = moduleNames.map((moduleName) => isVueFile(moduleName)
                ? moduleName + VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR
                : moduleName.endsWith('.vue?internal')
                    ? moduleName.replace(/\?internal$/, '') +
                        VIRTUAL_FILENAME_SEPARATOR +
                        INTERNAL_MODULE_SELECTOR
                    : moduleName);
            // TODO: Support paths mapped to .vue files, if needed.
            const result = resolveModuleNames != null
                ? resolveModuleNames(newModuleNames, containingFile, reusedNames, redirectedReferences, options)
                : [];
            result.forEach((resolved) => {
                if (resolved != null && isVirtualFile(resolved.resolvedFileName)) {
                    context.tryCreateScriptInfo(resolved.resolvedFileName); // Trigger .vue script info creation.
                }
            });
            return result;
        });
    });
}
function patchServiceHost(context) {
    patchWatchFile(context);
    patchWatchDirectory(context);
    patchFileExists(context);
    patchReadFile(context);
    patchReadDirectory(context);
}
function patchWatchFile(context) {
    tryPatchMethod(context.serviceHost, 'watchFile', (watchFile) => {
        context.log(`[patch] Override watchFile to watch virtual files. (ServiceHost)`);
        return wrapFn('watchFile', (fileName, callback, pollingInterval, options) => {
            if (isVirtualFile(fileName)) {
                return watchFile(getContainingFile(fileName), (_, event) => {
                    callback(fileName, event);
                }, pollingInterval, options);
            }
            return watchFile(fileName, callback, pollingInterval, options);
        });
    });
}
function patchWatchDirectory(context) {
    tryPatchMethod(context.serviceHost, 'watchDirectory', (watchDirectory) => {
        context.log(`[patch] Override watchFile to watch virtual files. (ServiceHost)`);
        return wrapFn('watchDirectory', (fileName, callback, recursive, options) => {
            return watchDirectory(fileName, (fileName) => {
                if (isVueFile(fileName)) {
                    // TODO: Ensure this does mess with fileName case-sensitivity
                    const document = context.store.get(fileName);
                    if (document != null) {
                        callback(document.getDocumentFileName('_module'));
                    }
                }
                else {
                    callback(fileName);
                }
            }, recursive, options);
        });
    });
}
function triggerFileUpdate(context, fileName) {
    const scriptInfo = context.projectService.getScriptInfo(fileName);
    if (scriptInfo != null) {
        // @ts-expect-error - internal method but it's better for performance compared to it's public counter part `reloadFromFile()`.
        scriptInfo.delayReloadNonMixedContentFile();
    }
}
function patchScriptInfo(context, scriptInfo) {
    if (scriptInfo == null)
        throw new Error('ScriptInfo is required.');
    tryPatchMethod(scriptInfo, 'editContent', (editContent) => {
        context.log(`[patch] Override editContent() of "${scriptInfo.fileName}" to sync virtual files. (ScriptInfo)`);
        return wrapFn('editContent', (start, end, newText) => {
            const document = context.store.get(scriptInfo.fileName);
            if (document == null)
                throw new Error('VueTextDocument should exist for every ScriptInfo.');
            const range = {
                start: document.positionAt(start),
                end: document.positionAt(end),
            };
            editContent(start, end, newText);
            VueTextDocument.update(document, [{ range, text: newText }], getLastNumberFromVersion(scriptInfo.getLatestVersion()))
                .all()
                .forEach((document) => triggerFileUpdate(context, document.fsPath));
        });
    });
}

function isNotNull(value) {
    return value != null;
}

function isArray(value) {
    return Array.isArray(value);
}
function first(items) {
    return items[0];
}
function last(items) {
    return items[items.length - 1];
}
function findNextSibling(items, item) {
    const index = items.indexOf(item);
    if (index >= 0)
        return items[index + 1];
}
const cacheStringFunction = (fn) => {
    const cache = Object.create(null);
    return ((str) => {
        const hit = cache[str];
        return hit !== null && hit !== void 0 ? hit : (cache[str] = fn(str));
    });
};
const camelizeRE = /[^A-Za-z0-9]([A-Za-z0-9])/g;
const camelize = cacheStringFunction((str) => {
    return str.replace(camelizeRE, (_, c) => typeof c === 'string' ? c.toUpperCase() : '');
});
const hyphenateRE = /\B([A-Z])/g;
const hyphenate = cacheStringFunction((str) => {
    return camelize(str).replace(hyphenateRE, '-$1').toLowerCase();
});
const capitalize = cacheStringFunction((str) => {
    return str.charAt(0).toUpperCase() + str.slice(1);
});
const pascalCase = cacheStringFunction((str) => capitalize(camelize(str)));
const kebabCase = hyphenate;

function getComponentName(fileName) {
    return pascalCase(Path.posix.basename(fileName).replace(/\.(vue|ts|tsx|js|jsx)$/, ''));
}

function createCachedAnalyzer() {
    const cache = new QuickLRU({ maxSize: 1000 });
    const analyzer = createFullAnalyzer([], {
        babel: { plugins: ['typescript', 'jsx'] },
    });
    return (document) => {
        var _a;
        const key = `${document.version}::${document.fsPath}`;
        const info = (_a = cache.get(key)) !== null && _a !== void 0 ? _a : analyzer.analyze(document.getText(), document.fsPath);
        cache.set(key, info);
        return info;
    };
}
// eslint-disable-next-line @typescript-eslint/explicit-function-return-type
function createServerHelper(context, languageService) {
    const getComponentInfo = createCachedAnalyzer();
    function isFeatureEnabled(featureName, checkFor = true) {
        const feature = context.config.features[featureName];
        return Array.isArray(feature)
            ? feature.includes(checkFor)
            : feature === true || feature === checkFor;
    }
    /**
     * Find node in template containing given position.
     * @param fileName Vue or Virtual Block filename
     * @param position Position or range
     */
    function findTemplateNodeAtPosition(fileName, position) {
        const document = getRenderDoc(fileName);
        if ((document === null || document === void 0 ? void 0 : document.ast) != null) {
            if (typeof position === 'number') {
                return {
                    ...findTemplateNodeAt(document.ast, position),
                    document,
                };
            }
            else {
                return {
                    ...findTemplateNodeInRange(document.ast, position.pos, position.end),
                    document,
                };
            }
        }
        return { node: null, ancestors: [], document: null };
    }
    function findTemplateChildren(fileName, position) {
        const document = getRenderDoc(fileName);
        return (document === null || document === void 0 ? void 0 : document.ast) != null
            ? typeof position === 'number'
                ? findTemplateChildrenInRange(document.ast, position, position)
                : findTemplateChildrenInRange(document.ast, position.pos, position.end)
            : [];
    }
    function getDocument(fileName) {
        var _a, _b;
        return isVirtualFile(fileName)
            ? (_b = (_a = context.store.get(getContainingFile(fileName))) === null || _a === void 0 ? void 0 : _a.getDocument(fileName)) !== null && _b !== void 0 ? _b : null : context.store.get(fileName);
    }
    function getVueDocument(fileName) {
        return isVueFile(fileName)
            ? context.store.get(fileName)
            : isVirtualFile(fileName)
                ? context.store.get(getContainingFile(fileName))
                : null;
    }
    function getBlockAt(fileName, position) {
        var _a;
        const document = getVueDocument(fileName);
        if (document != null) {
            if (typeof position === 'number') {
                return (_a = document.blockAt(position)) !== null && _a !== void 0 ? _a : null;
            }
            else {
                const a = document.blockAt(position.pos);
                const b = document.blockAt(position.end);
                if (a === b)
                    return a !== null && a !== void 0 ? a : null;
            }
        }
        return null;
    }
    function getDocumentForBlock(fileName, block) {
        const document = getVueDocument(fileName);
        if (document != null) {
            if (block.type === 'template') {
                return document.getDocument(RENDER_SELECTOR);
            }
            const selector = document.getBlockSelector(block);
            if (selector != null) {
                return document.getDocument(selector);
            }
        }
        return null;
    }
    function getDocumentAt(fileName, position) {
        const block = getBlockAt(fileName, position);
        if (block != null) {
            return getDocumentForBlock(fileName, block);
        }
        return null;
    }
    function isRenderFunctionDocument(document) {
        var _a;
        return ((_a = document) === null || _a === void 0 ? void 0 : _a.selector.type) === RENDER_SELECTOR;
    }
    function isRenderFunctionFileName(fileName) {
        var _a;
        return (isVirtualFile(fileName) &&
            ((_a = parseVirtualFileName(fileName)) === null || _a === void 0 ? void 0 : _a.selector.type) === RENDER_SELECTOR);
    }
    function isVueModuleFileName(fileName) {
        var _a;
        return (isVirtualFile(fileName) &&
            ((_a = parseVirtualFileName(fileName)) === null || _a === void 0 ? void 0 : _a.selector.type) === MODULE_SELECTOR);
    }
    function getRenderDoc(fileName) {
        var _a;
        return (_a = getVueDocument(fileName)) === null || _a === void 0 ? void 0 : _a.getDocument('_render');
    }
    function getTextSpan(document, span, node) {
        if (isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result != null)
                return {
                    start: result.offset,
                    length: Math.min(result.length, span.length),
                };
        }
        return span;
    }
    function getResolvedModule(fileName, importSource) {
        const program = languageService.getProgram();
        const sourceFile = program === null || program === void 0 ? void 0 : program.getSourceFile(fileName);
        const resolvedModules = sourceFile === null || sourceFile === void 0 ? void 0 : sourceFile.resolvedModules;
        return resolvedModules === null || resolvedModules === void 0 ? void 0 : resolvedModules.get(importSource);
    }
    function findTypeScriptNodeAtPosition(fileName, position) {
        var _a;
        const findNodeAt = (start, end, container) => {
            var _a;
            const node = container
                .getChildren()
                .find((node) => node.pos <= start && end <= node.end);
            if (node != null) {
                if (node.pos === start && node.pos === end)
                    return node;
                else
                    return (_a = findNodeAt(start, end, node)) !== null && _a !== void 0 ? _a : node;
            }
            else {
                return container;
            }
        };
        const scriptFile = (_a = languageService.getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(fileName);
        if (scriptFile != null) {
            const result = typeof position === 'number'
                ? findNodeAt(position, position, scriptFile)
                : findNodeAt(position.pos, position.end, scriptFile);
            if (result !== scriptFile)
                return result;
        }
    }
    return {
        isFeatureEnabled,
        findTemplateChildren,
        findTemplateNodeAtPosition,
        findTypeScriptNodeAtPosition,
        getComponentInfo,
        getDocument,
        getDocumentAt,
        getRenderDoc,
        getTextSpan,
        getVueDocument,
        isRenderFunctionDocument,
        isRenderFunctionFileName,
        isVueModuleFileName,
        getResolvedModule,
    };
}
function computeIdentifierReplacement(source, identifer) {
    const RE = new RegExp(`\\b${identifer}\\b`);
    const match = RE.exec(source);
    if (match == null)
        return { prefixText: source, suffixText: '' };
    return {
        prefixText: source.substr(0, match.index),
        suffixText: source.substr(match.index + match[0].length),
    };
}
function getPaddingLength(source, offset = 0) {
    source = source.substr(offset);
    const match = /^[\s\r\n]+/m.exec(source);
    return match != null ? match[0].length : 0;
}
function getFilenameForNewComponent(context, directory, usedNames = new Set()) {
    let name = 'Component0';
    let fileName = Path.posix.join(directory, 'Component0.vue');
    let index = 0;
    if (context.serviceHost.directoryExists(directory)) {
        const files = new Set(context.serviceHost.readDirectory(directory, ['.vue'], [], [], 1));
        while (files.has(fileName) || usedNames.has(name)) {
            name = `Component${++index}`;
            fileName = Path.posix.join(directory, `${name}.vue`);
        }
    }
    return fileName;
}
function indent(code, size = 2, prefix = true) {
    const text = ' '.repeat(size);
    return (prefix ? text : '') + code.split('\n').join('\n' + text);
}
function getCode(...lines) {
    return lines.flat().join('\n');
}
function isSpanInSourceRange(span, range) {
    if (range == null)
        return false;
    return (range.start.offset <= span.start &&
        span.start + span.length <= range.end.offset);
}
function isTextRangeInSourceRange(textRange, range) {
    if (range == null)
        return false;
    const span = typeof textRange === 'number'
        ? { start: textRange, length: 0 }
        : { start: textRange.pos, length: textRange.end - textRange.pos };
    return isSpanInSourceRange(span, range);
}
function getScriptFileName(document) {
    if (document.descriptor.scriptSetup != null) {
        return document.getDocumentFileName(SCRIPT_SETUP_BLOCK_SELECTOR);
    }
    else {
        return document.getDocumentFileName(SCRIPT_BLOCK_SELECTOR);
    }
}

function registerLocalComponentWithSource(document, componentInfo, source, preferences, importStatement) {
    const { scriptSetup, script } = document.descriptor;
    const changes = [];
    importStatement = importStatement !== null && importStatement !== void 0 ? importStatement : `import ${source.exportName != null
        ? source.localName === source.exportName
            ? `{ ${source.exportName} }`
            : `{ ${source.exportName} as ${source.localName} }`
        : source.localName} from ${JSON.stringify(source.moduleName)}`;
    if (scriptSetup != null) {
        changes.push({
            span: { start: scriptSetup.loc.start.offset + 1, length: 0 },
            newText: importStatement,
        });
    }
    else if (script != null) {
        changes.push({
            span: { start: script.loc.start.offset + 1, length: 0 },
            newText: importStatement,
        });
        changes.push(...registerComponentAPI(document, componentInfo, 'components', source.localName, source.localName).changes);
    }
    else {
        changes.push({
            span: { start: 0, length: 0 },
            newText: preferences.mode === 'normal'
                ? [
                    preferences.language === 'js'
                        ? `<script>`
                        : `<script lang=${JSON.stringify(preferences.language)}>`,
                    `import { defineComponent } from 'vue'`,
                    importStatement,
                    '',
                    `export default defineComponent({`,
                    `  components: { ${source.localName} }`,
                    `})`,
                    `</script>`,
                    '\n',
                ].join('\n')
                : [
                    preferences.language === 'js'
                        ? `<script setup>`
                        : `<script lang=${JSON.stringify(preferences.language)} setup>`,
                    importStatement,
                    `</script>`,
                    '\n',
                ].join('\n'),
        });
    }
    return changes;
}
function registerComponentAPI(document, componentInfo, apiName, name, value, preferences) {
    if (apiName === 'setupComputed' ||
        apiName === 'setupMethods' ||
        apiName === 'data') {
        return addOptionToConfigFunction(document, componentInfo, apiName, name, value);
    }
    else {
        return addOptionToConfigObject(document, componentInfo, apiName, name, value, preferences);
    }
}
function addOptionToConfigFunction(document, componentInfo, apiName, name, value, preferences) {
    const { scriptSetup, script } = document.descriptor;
    if (scriptSetup != null) {
        return addOptionToScriptSetupConfig(scriptSetup, componentInfo.scriptSetup, apiName.startsWith('setup') ? apiName.substr(5).toLowerCase() : apiName, name, value);
    }
    else if (script != null) {
        const changes = [];
        let scriptStartOffset;
        if (apiName === 'data') ;
        else {
            const api = apiName.substr(5).toLowerCase();
            let code;
            if (api === 'computed') {
                const hasComputed = /\bcomputed\b/.test(script.content);
                if (!hasComputed) {
                    changes.push({
                        newText: getCode(`import { computed } from 'vue'`, ``),
                        span: {
                            start: script.loc.start.offset +
                                getPaddingLength(script.content, script.loc.start.offset),
                            length: 0,
                        },
                    });
                }
                code = `const ${name} = computed(${value})`;
            }
            else if (api === 'methods') {
                code = `function ${name}${value.replace(/^function /, '')}`;
            }
            else {
                code = `const ${name} = ${value}`;
            }
            if (componentInfo.fnSetupOption != null) {
                if (componentInfo.fnSetupOption.return != null) {
                    const loc = componentInfo.fnSetupOption.return.loc;
                    const start = componentInfo.fnSetupOption.loc.start.offset +
                        componentInfo.fnSetupOption.loc.source.lastIndexOf('return');
                    changes.push({
                        newText: getCode(indent(code, 4, false), ``, `    `),
                        span: { start: start, length: 0 },
                    });
                    if (isSimpleIdentifier(loc.source)) {
                        changes.push({
                            newText: getCode(`{`, `    ...${loc.source.trim()}`, `    ${name},`, `}`),
                            span: { start: loc.start.offset, length: loc.source.length },
                        });
                    }
                    else {
                        if (loc.source.includes('\n')) {
                            changes.push({
                                newText: getCode(``, `      ${name},`),
                                span: {
                                    start: loc.start.offset + 1,
                                    length: 0,
                                },
                            });
                        }
                        else {
                            changes.push({
                                newText: getCode(` ${name},`),
                                span: {
                                    start: loc.start.offset + 1,
                                    length: 0,
                                },
                            });
                        }
                    }
                }
            }
            else {
                const newText = getCode(`function (props, context) {`, `  ${code}`, `  return { ${name} }`, `}`);
                const result = addOptionToScriptConfig(script, componentInfo.options, componentInfo.fnSetupOption, '', 'setup', newText);
                changes.push(...result.changes);
            }
        }
        return {
            scriptStartOffset,
            changes,
        };
    }
    else {
        return {
            changes: [],
        };
    }
}
function addOptionToConfigObject(document, componentInfo, apiName, name, value, preferences) {
    const { scriptSetup, script } = document.descriptor;
    if (scriptSetup != null) {
        return addOptionToScriptSetupConfig(scriptSetup, componentInfo.scriptSetup, apiName, name, value);
    }
    else if (script != null) {
        return addOptionToScriptConfig(script, componentInfo.options, componentInfo.fnSetupOption, apiName, name, value);
    }
    else if (preferences != null) {
        const changes = [];
        const language = preferences.language === 'js' ? '' : ` lang="${preferences.language}"`;
        const mode = preferences.mode === 'normal' ? '' : ' setup';
        const startTag = `<script${language}${mode}>\n`;
        const result = preferences.mode === 'normal'
            ? addOptionToScriptConfig(createFakeScriptBock(preferences.language), undefined, undefined, apiName, name, value)
            : addOptionToScriptSetupConfig(createFakeScriptSetupBock(preferences.language), undefined, apiName, name, value);
        document.markDirty(); // <- potentially changed
        changes.push({
            newText: startTag,
            span: { start: 0, length: 0 },
        });
        changes.push(...result.changes);
        changes.push({
            newText: getCode(`</script>`, ``, ``),
            span: { start: 0, length: 0 },
        });
        return {
            changes,
            renameLocation: result.renameLocation != null
                ? result.renameLocation + startTag.length
                : 0,
            scriptStartOffset: startTag.length,
        };
    }
    return { changes: [] };
}
function createFakeScriptBock(lang) {
    if (lang === 'js')
        lang = undefined;
    return {
        content: '',
        attrs: lang != null ? { lang } : {},
        lang,
        type: 'script',
        loc: {
            source: '',
            start: {
                offset: 0,
                line: 0,
                column: 0,
            },
            end: {
                offset: 0,
                line: 0,
                column: 0,
            },
        },
    };
}
function createFakeScriptSetupBock(lang) {
    if (lang === 'js')
        lang = undefined;
    return {
        content: '',
        attrs: lang != null ? { lang, setup: true } : { setup: true },
        lang,
        setup: true,
        type: 'script',
        loc: {
            source: '',
            start: {
                offset: 0,
                line: 0,
                column: 0,
            },
            end: {
                offset: 0,
                line: 0,
                column: 0,
            },
        },
    };
}
function addOptionToScriptConfig(script, optionsLoc, fnSetupOptionLoc, apiName, name, value) {
    const changes = [];
    let renameLocation;
    if (apiName === '') {
        const code = genPropertyCode(name, value, 2);
        if (optionsLoc != null) {
            const start = optionsLoc.loc.start.offset + 1;
            const newText = getCode(``, `  ${code}`);
            renameLocation = start + newText.indexOf(code);
            changes.push({ newText, span: { start: start, length: 0 } });
        }
        else if (fnSetupOptionLoc != null) {
            const setupLoc = fnSetupOptionLoc.loc;
            changes.push({
                newText: getCode('{', code, '  setup: '),
                span: { start: setupLoc.start.offset, length: 0 },
            }, {
                newText: `\n}`,
                span: { start: setupLoc.end.offset, length: 0 },
            });
        }
        else {
            const hasDefineComponent = /\bdefineComponent\b/.test(script.content);
            if (!hasDefineComponent) {
                const padding = getPaddingLength(script.content);
                changes.push({
                    newText: getCode(`import { defineComponent } from 'vue';`, ''),
                    span: { start: script.loc.start.offset + padding, length: 0 },
                });
            }
            const newText = getCode('', `export default defineComponent({`, code, `})`, '');
            changes.push({
                span: { start: script.loc.end.offset, length: 0 },
                newText: newText,
            });
        }
        return { changes };
    }
    const code = genPropertyCode(name, value, 4);
    if (optionsLoc != null) {
        const api = optionsLoc.properties[apiName];
        if (isSimpleIdentifier(optionsLoc.loc.source)) {
            const start = optionsLoc.loc.start.offset;
            const comma = optionsLoc.loc.source.trimEnd().endsWith(',') ? '' : ',';
            const newText = getCode(`{`, `  ...${optionsLoc.loc.source}${comma}`, `  ${code}`, '},');
            renameLocation = start + newText.indexOf(code);
            changes.push({
                newText,
                span: { start: start, length: optionsLoc.loc.source.length },
            });
        }
        else if (api != null) {
            if (isSimpleIdentifier(api.loc.source)) {
                const start = api.loc.start.offset;
                const comma = api.loc.source.trimEnd().endsWith(',') ? '' : ',';
                const newText = getCode(`{`, `    ...${api.loc.source}${comma}`, `    ${code}`, '},');
                renameLocation = start + newText.indexOf(code);
                changes.push({
                    newText,
                    span: { start: start, length: api.loc.source.length },
                });
            }
            else {
                const start = api.loc.start.offset + 1;
                const newText = getCode(``, `    ${code}`);
                renameLocation = start + newText.indexOf(code);
                changes.push({ newText, span: { start: start, length: 0 } });
            }
        }
        else {
            const newText = getCode(`${apiName}: {`, `    ${code}`, `  },`, '  ');
            const padding = getPaddingLength(optionsLoc.loc.source, 1);
            const start = optionsLoc.loc.start.offset + padding + 1;
            renameLocation = start + padding + newText.indexOf(code);
            changes.push({ newText, span: { start, length: 0 } });
        }
    }
    else if (fnSetupOptionLoc != null) {
        const setupLoc = fnSetupOptionLoc.loc;
        const newText = getCode('{', `  ${apiName}: {`, `    ${code}`, '  },', '  setup: ');
        renameLocation = setupLoc.start.offset + newText.indexOf(code);
        changes.push({
            newText,
            span: { start: setupLoc.start.offset, length: 0 },
        }, {
            newText: `\n}`,
            span: { start: setupLoc.end.offset, length: 0 },
        });
    }
    else {
        const hasDefineComponent = /\bdefineComponent\b/.test(script.content);
        if (!hasDefineComponent) {
            const padding = getPaddingLength(script.content, script.loc.start.offset);
            changes.push({
                newText: getCode(`import { defineComponent } from 'vue';`, ''),
                span: { start: script.loc.start.offset + padding, length: 0 },
            });
        }
        const newText = getCode('', `export default defineComponent({`, `  ${apiName}: {`, `    ${code}`, `  },`, `})`, '');
        renameLocation = script.loc.end.offset + newText.indexOf(code);
        changes.push({
            span: { start: script.loc.end.offset, length: 0 },
            newText: newText,
        });
    }
    return { changes, renameLocation };
}
function addOptionToScriptSetupConfig(scriptSetup, scriptSetupLoc, apiName, name, value) {
    let renameLocation;
    const changes = [];
    if (apiName === 'components' || apiName === 'directives') ;
    else if (apiName === 'props') {
        if ((scriptSetupLoc === null || scriptSetupLoc === void 0 ? void 0 : scriptSetupLoc.defineProps) != null) {
            const props = scriptSetupLoc.defineProps;
            const padding = getPaddingLength(props.loc.source, props.loc.source.indexOf('{'));
            changes.push({
                span: { start: props.loc.start.offset + padding, length: 0 },
                newText: getCode(`  ${genPropertyCode(name, value, 2)}`, ``),
            });
        }
        else {
            changes.push({
                span: {
                    start: scriptSetup.loc.start.offset +
                        getPaddingLength(scriptSetup.content, scriptSetup.loc.start.offset),
                    length: 0,
                },
                newText: getCode(`import { defineProps } from 'vue'`, `const props = defineProps({`, `  ${genPropertyCode(name, value, 2)}`, `})`, ``),
            });
        }
    }
    else if (apiName === 'emits') {
        if ((scriptSetupLoc === null || scriptSetupLoc === void 0 ? void 0 : scriptSetupLoc.defineEmit) != null) {
            const emit = scriptSetupLoc.defineEmit;
            const padding = getPaddingLength(emit.loc.source, emit.loc.source.indexOf('{'));
            changes.push({
                span: { start: emit.loc.start.offset + padding, length: 0 },
                newText: getCode(`  ${genPropertyCode(name, value, 2)}`, ``),
            });
        }
        else {
            changes.push({
                span: {
                    start: scriptSetup.loc.start.offset +
                        getPaddingLength(scriptSetup.content, scriptSetup.loc.start.offset),
                    length: 0,
                },
                newText: getCode(`import { defineEmit } from 'vue'`, `const emit = defineEmit({`, `  ${genPropertyCode(name, value, 2)}`, `})`, ``),
            });
        }
    }
    else if (apiName === 'watch') {
        const hasWatch = /\bwatch\b/.test(scriptSetup.content);
        if (!hasWatch) {
            changes.push({
                span: {
                    start: scriptSetup.loc.start.offset +
                        getPaddingLength(scriptSetup.content, scriptSetup.loc.start.offset),
                    length: 0,
                },
                newText: getCode(`import { watch } from 'vue'`, ``),
            });
        }
        changes.push({
            span: { start: scriptSetup.loc.end.offset, length: 0 },
            newText: getCode(`watch(() => ${name}, ${value})`, ``),
        });
    }
    else if (apiName === 'methods') {
        const newText = getCode(value.startsWith('function (')
            ? value.replace(/^function /, `function ${name}`)
            : `const ${name} = ${value}`, '');
        renameLocation = scriptSetup.loc.end.offset + newText.indexOf(name);
        changes.push({
            span: { start: scriptSetup.loc.end.offset, length: 0 },
            newText,
        });
    }
    else if (apiName === 'computed') {
        const hasComputed = /\bcomputed\b/.test(scriptSetup.content);
        if (!hasComputed) {
            changes.push({
                span: {
                    start: scriptSetup.loc.start.offset +
                        getPaddingLength(scriptSetup.content, scriptSetup.loc.start.offset),
                    length: 0,
                },
                newText: getCode(`import { computed } from 'vue'`, ``),
            });
        }
        renameLocation = scriptSetup.loc.end.offset + 6; // < length of 'const '
        changes.push({
            span: { start: scriptSetup.loc.end.offset, length: 0 },
            newText: getCode(`const ${name} = computed(${value})`, ``),
        });
    }
    else if (apiName === 'data') {
        const hasRef = /\bref\b/.test(scriptSetup.content);
        if (!hasRef) {
            changes.push({
                span: {
                    start: scriptSetup.loc.start.offset +
                        getPaddingLength(scriptSetup.content, scriptSetup.loc.start.offset),
                    length: 0,
                },
                newText: getCode(`import { ref } from 'vue'`, ``),
            });
        }
        renameLocation = scriptSetup.loc.end.offset + 6; // < length of 'const '
        changes.push({
            span: { start: scriptSetup.loc.end.offset, length: 0 },
            newText: getCode(`const ${name} = ref(${value})`, ``),
        });
    }
    return { changes, renameLocation };
}
function genPropertyCode(name, value, indentSize) {
    if (name === value) {
        return `${name},`;
    }
    else if (value.trim().startsWith('function (')) {
        return `${name}${indent(value.replace(/^function /, ''), indentSize, false)},`;
    }
    else {
        return `${name}: ${indent(value, indentSize, false)},`;
    }
}

const noop = {
    toggleLineComment(fileName, textRange) {
        return [];
    },
    toggleMultilineComment(fileName, textRange) {
        return [];
    },
    commentSelection(fileName, textRange) {
        return [];
    },
    uncommentSelection(fileName, textRange) {
        return [];
    },
    getRenameInfo(fileName, position, options) {
        return {
            canRename: false,
            localizedErrorMessage: 'Not implemented.',
        };
    },
    findRenameLocations(fileName, position, findInStrings, findInComments) {
        return undefined;
    },
    getApplicableRefactors(fileName, positionOrRange, preferences) {
        return [];
    },
    getBraceMatchingAtPosition(fileName, position) {
        return [];
    },
    getBreakpointStatementAtPosition(fileName, position) {
        return undefined;
    },
    getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
        return [];
    },
    getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
        return {
            changes: [],
        };
    },
    getCompilerOptionsDiagnostics() {
        return [];
    },
    getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
        return undefined;
    },
    getCompletionEntrySymbol(fileName, position, name, source) {
        return undefined;
    },
    getCompletionsAtPosition(fileName, position, options) {
        return undefined;
    },
    getDefinitionAndBoundSpan(fileName, position) {
        return undefined;
    },
    getDefinitionAtPosition(fileName, position) {
        return undefined;
    },
    getDocCommentTemplateAtPosition(fileName, position) {
        return undefined;
    },
    getDocumentHighlights(fileName, position, filesToSearch) {
        return undefined;
    },
    getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
        return [];
    },
    getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
        return undefined;
    },
    getEmitOutput(fileName) {
        return { outputFiles: [], emitSkipped: true };
    },
    getEncodedSemanticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getEncodedSyntacticClassifications(fileName, span) {
        return {
            spans: [],
            endOfLineState: 0,
        };
    },
    getFormattingEditsAfterKeystroke(fileName, position, key, options) {
        return [];
    },
    getFormattingEditsForDocument(fileName, options) {
        return [];
    },
    getFormattingEditsForRange(fileName, start, end, options) {
        return [];
    },
    getImplementationAtPosition(fileName, position) {
        return undefined;
    },
    getIndentationAtPosition(fileName, position, options) {
        return 0;
    },
    getJsxClosingTagAtPosition(fileName, position) {
        return undefined;
    },
    getNameOrDottedNameSpan(fileName, startPos, endPos) {
        return undefined;
    },
    getNavigateToItems(searchValue) {
        return [];
    },
    getNavigationBarItems(fileName) {
        return [];
    },
    getNavigationTree(fileName) {
        return {
            text: '',
            kind: 'module',
            kindModifiers: 'component',
            spans: [],
            nameSpan: undefined,
        };
    },
    getOccurrencesAtPosition(fileName, position) {
        return undefined;
    },
    getOutliningSpans(fileName) {
        return [];
    },
    getQuickInfoAtPosition(fileName, position) {
        return undefined;
    },
    getReferencesAtPosition(fileName, position) {
        return undefined;
    },
    getSemanticClassifications(fileName, span) {
        return [];
    },
    getSemanticDiagnostics(fileName) {
        return [];
    },
    getSignatureHelpItems(fileName, position, options) {
        return undefined;
    },
    getSmartSelectionRange(fileName, position) {
        return {
            textSpan: { start: position, length: 0 },
        };
    },
    getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
        return undefined;
    },
    getSuggestionDiagnostics(fileName) {
        return [];
    },
    getSyntacticClassifications(fileName, span) {
        return [];
    },
    getSyntacticDiagnostics(fileName) {
        return [];
    },
    getTodoComments(fileName, descriptors) {
        return [];
    },
    getTypeDefinitionAtPosition(fileName, position) {
        return [];
    },
    findReferences(fileName, position) {
        return undefined;
    },
    prepareCallHierarchy(fileName, position) {
        return undefined;
    },
    provideCallHierarchyIncomingCalls(fileName, position) {
        return [];
    },
    provideCallHierarchyOutgoingCalls(fileName, position) {
        return [];
    },
    isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
        return false;
    },
    organizeImports(scope, formatOptions, preferences) {
        return [];
    },
    getProgram() {
        throw new Error('Not Implemented');
    },
    cleanupSemanticCache() {
        throw new Error('Not Implemented');
    },
    dispose() {
        throw new Error('Not Implemented');
    },
    applyCodeActionCommand() {
        return Promise.reject('Not Implemented');
    },
};

function createVirtualLanguageServer(config) {
    function removePrefix(fileName) {
        return fileName.startsWith('^vue:') ? fileName.substr(5) : fileName;
    }
    function choose(fileName) {
        var _a, _b;
        return ((_b = (_a = config.context.projectService
            .getDefaultProjectForFile(config.context.typescript.server.toNormalizedPath(fileName), false)) === null || _a === void 0 ? void 0 : _a.getLanguageService()) !== null && _b !== void 0 ? _b : config.service);
    }
    return wrapInTrace('VirtualLanguageServer', {
        ...noop,
        getQuickInfoAtPosition(rawFileName, position) {
            const fileName = removePrefix(rawFileName);
            return choose(fileName).getQuickInfoAtPosition(fileName, position);
        },
    });
}

function encode(value) {
    return JSON.stringify(value);
}
function decode(value) {
    return JSON.parse(value);
}
const REFACTORS = {
    EXTRACT_COMPONENT: 'vue:extract-component',
    EXTRACT_METHOD: 'vue:extract-method',
    EXTRACT_COMPUTED: 'vue:extract-computed',
    MOVE_DIRECTIVE_TO_COMPONENT: 'vue:move-directive-to-component',
    ADD_TO_SETUP_CONTEXT: 'vue:add-to-setup-context',
};

const AddToSetupContextRefactor = {
    version: '*',
    name: REFACTORS.ADD_TO_SETUP_CONTEXT,
    findRefactors({ helpers, context, service }, fileName, position, preferences) {
        var _a, _b;
        const addToSetupContext = {
            name: 'default',
            description: 'Add to return statement',
        };
        const refactor = {
            name: REFACTORS.ADD_TO_SETUP_CONTEXT,
            description: 'Add to return statement',
            inlineable: true,
            actions: [addToSetupContext],
        };
        const document = helpers.getVueDocument(fileName);
        if (document != null) {
            const info = helpers.getComponentInfo(document);
            context.log(`@@DEBUG: Is position in setup function?`);
            if (((_a = info.fnSetupOption) === null || _a === void 0 ? void 0 : _a.return) != null &&
                isTextRangeInSourceRange(position, info.fnSetupOption.loc) &&
                !isTextRangeInSourceRange(position, info.fnSetupOption.return.loc)) {
                const node = helpers.findTypeScriptNodeAtPosition(fileName, position);
                if (node != null && context.typescript.isIdentifier(node)) {
                    const name = node.text;
                    if (((_b = info.identifierSource[name]) === null || _b === void 0 ? void 0 : _b.name) === 'setup') {
                        addToSetupContext.notApplicableReason =
                            'Already exists in return statement';
                        return [refactor];
                    }
                    const parent = helpers.findTypeScriptNodeAtPosition(fileName, {
                        pos: info.fnSetupOption.loc.start.offset,
                        end: info.fnSetupOption.loc.end.offset,
                    });
                    if (parent != null) {
                        if (context.typescript.isFunctionExpression(parent) ||
                            context.typescript.isArrowFunction(parent) ||
                            context.typescript.isMethodDeclaration(parent)) {
                            const program = service.getProgram();
                            if (parent.body != null && program != null) {
                                const checker = program.getTypeChecker();
                                const symbols = checker.getSymbolsInScope(parent.body, context.typescript.SymbolFlags.BlockScopedVariable);
                                if (symbols.some((sym) => sym.name === name)) {
                                    addToSetupContext.description = `Add '${name}' to return statement`;
                                    return [refactor];
                                }
                            }
                        }
                    }
                }
            }
        }
        addToSetupContext.notApplicableReason = 'Could not find identifier';
        return [refactor];
    },
    applyRefactor({ helpers, context, service }, fileName, _, position, refactorName, action) {
        var _a;
        if (refactorName !== REFACTORS.ADD_TO_SETUP_CONTEXT)
            return;
        const changes = [];
        const document = helpers.getVueDocument(fileName);
        if (document != null) {
            const info = helpers.getComponentInfo(document);
            if (((_a = info.fnSetupOption) === null || _a === void 0 ? void 0 : _a.return) != null) {
                const node = helpers.findTypeScriptNodeAtPosition(fileName, position);
                if (node != null && context.typescript.isIdentifier(node)) {
                    const loc = info.fnSetupOption.return.loc;
                    if (isSimpleIdentifier(loc.source)) {
                        changes.push({
                            newText: getCode(`{`, `    ...${loc.source.trim()}`, `    ${node.text},`, `}`),
                            span: { start: loc.start.offset, length: loc.source.length },
                        });
                    }
                    else {
                        if (loc.source.includes('\n')) {
                            changes.push({
                                newText: getCode(``, `      ${node.text},`),
                                span: {
                                    start: loc.start.offset + 1,
                                    length: 0,
                                },
                            });
                        }
                        else {
                            changes.push({
                                newText: getCode(` ${node.text},`),
                                span: {
                                    start: loc.start.offset + 1,
                                    length: 0,
                                },
                            });
                        }
                    }
                }
            }
            return {
                edits: [
                    {
                        fileName: document.fsPath,
                        textChanges: changes,
                    },
                ],
            };
        }
    },
};

const RefactorExtractComponent = {
    version: '*',
    name: REFACTORS.EXTRACT_COMPONENT,
    findRefactors({ helpers, context }, fileName, position) {
        const extractInCurrentDirectory = {
            name: encode({ directory: '.' }),
            description: 'Extract to component in current directory',
        };
        const refactor = {
            name: REFACTORS.EXTRACT_COMPONENT,
            description: 'Extract to component',
            inlineable: true,
            actions: [extractInCurrentDirectory],
        };
        // Check if refactor can be applied
        const document = helpers.getDocumentAt(fileName, position);
        if (helpers.isRenderFunctionDocument(document)) {
            const nodes = findNodes(document, position);
            if (nodes.length === 0) {
                extractInCurrentDirectory.notApplicableReason = 'No active selection';
            }
            else if (document.ast != null && nodes.length > 1) {
                const parent = findParentNode(document.ast, nodes);
                if (isElementNode(parent)) {
                    if (detectConditionType(parent, nodes) === 'partial') {
                        extractInCurrentDirectory.notApplicableReason =
                            'Partial v-if/v-else-if/v-else tags cannot be extracted';
                    }
                }
            }
        }
        else {
            extractInCurrentDirectory.notApplicableReason =
                'Only supported in template block';
        }
        // Load project component directories.
        const project = context.getVueProjectForFile(fileName, true);
        project.config.preferences.componentsDirectories.forEach((directory) => {
            refactor.actions.push({
                name: encode({ directory }),
                description: `Extract to component in "${directory}" directory`,
                notApplicableReason: extractInCurrentDirectory.notApplicableReason,
            });
        });
        // TODO: Add options to extract including v-if/v-for on root node.
        if (refactor.actions.length > 4) {
            refactor.inlineable = false;
        }
        return [refactor];
    },
    applyRefactor(config, fileName, _, position, refactorName, actionName) {
        if (refactorName !== REFACTORS.EXTRACT_COMPONENT)
            return;
        const { helpers, context } = config;
        const args = decode(actionName);
        const document = helpers.getRenderDoc(fileName);
        if ((document === null || document === void 0 ? void 0 : document.ast) == null)
            return;
        // Resolved target directory.
        const project = context.getVueProjectForFile(fileName, true);
        const directoryName = args.directory === '.'
            ? Path.posix.dirname(fileName)
            : Path.posix.resolve(project.rootDir, args.directory);
        // Pick a placeholder component name.
        const info = helpers.getComponentInfo(document.container);
        const componentFileName = getFilenameForNewComponent(config.context, directoryName, new Set(info.components.map((component) => component.name)));
        const nodes = findNodes(document, position);
        // Known identifiers.
        const rawIdentifiers = new Set(nodes.flatMap((node) => node.scope.identifiers));
        // Collection of nodes that should be replaced when generating code of new component.
        const replaceNodes = new Map();
        const addSkipNodes = (skipped, all) => {
            skipped.forEach((node) => {
                replaceNodes.set(node, null);
                node.scope.identifiers.forEach((id) => rawIdentifiers.delete(id));
            });
            all.forEach((prop) => {
                if (!replaceNodes.has(prop)) {
                    prop.scope.identifiers.forEach((id) => rawIdentifiers.add(id));
                }
            });
        };
        // Detect how to process conditional directives.
        const parent = findParentNode(document.ast, nodes);
        const conditionType = detectConditionType(parent, nodes); // TODO: Provide choice to user.
        if (conditionType === 'preserve') {
            nodes.forEach((node) => {
                if (isElementNode(node)) {
                    addSkipNodes(node.props.filter((prop) => isDirectiveNode(prop) && /^(if|else-if|else)$/.test(prop.name)), [...node.props, ...node.children]);
                }
            });
        }
        // For single element node extracts, for look can be preserved.
        if (nodes.length === 1) {
            const el = first(nodes);
            if (isElementNode(el)) {
                const forNode = el.props.find((prop) => isDirectiveNode(prop) && prop.name === 'for');
                if (forNode != null) {
                    addSkipNodes([forNode], [...el.props, ...el.children]);
                }
            }
        }
        // Find components, v-model and v-emit.
        const components = new Set();
        const models = new Set();
        const emits = new Map();
        nodes.forEach((node) => traverseFast(node, (node) => {
            if (isComponentNode(node))
                components.add(node.tag);
            else if (isDirectiveNode(node)) {
                if (isSimpleExpressionNode(node.exp)) {
                    if (node.name === 'model') {
                        node.exp.scope.globals.forEach((identifier) => {
                            models.add(identifier);
                            rawIdentifiers.delete(identifier);
                        });
                    }
                    else if (node.name === 'on') {
                        if (isSimpleIdentifier(node.exp.content.trim())) {
                            const name = node.exp.content.trim();
                            addSkipNodes([node.exp], []);
                            replaceNodes.set(node.exp, createSimpleExpression(`$emit('${name}')`, false));
                            emits.set(name, name);
                        }
                    }
                }
            }
        }));
        // Generate template for new component
        const identifiers = Array.from(rawIdentifiers);
        const script = [];
        if (project.config.preferences.script.mode === 'normal') {
            // Generate options for new component
            const options = {
                components: genComponentsOptionsText({
                    info,
                    addImport: (text) => script.push(text),
                    fileName,
                    components,
                    componentFileName,
                }),
                props: genPropsOptionsText({ identifiers, models }),
                emits: genEmitsOptionsText({ emits, models }),
            };
            const optionsText = Array.from(Object.values(options))
                .filter(Boolean)
                .map((line) => '  ' + line)
                .join('\n');
            if (optionsText !== '') {
                script.unshift(`import { defineComponent } from 'vue'`);
                script.push(`export default defineComponent({\n${optionsText}\n})`);
            }
        }
        else {
            // We don't need to register components in <script setup> mode, so reusing
            // components option generation but rejecting output and keeping only imports.
            genComponentsOptionsText({
                info,
                addImport: (text) => script.push(text),
                fileName,
                components,
                componentFileName,
            });
            // TODO: Generate props/emits in object syntax with types maybe.
            const vueImports = [];
            if (identifiers.length > 0 || models.size > 0) {
                vueImports.push('defineProps');
                script.push(`const props = defineProps(${genProps(context, [
                    ...identifiers,
                    ...models,
                ])})`);
            }
            if (emits.size > 0 || models.size > 0) {
                vueImports.push('defineEmit');
                const text = genEmits(context, Array.from(emits.keys()), Array.from(models));
                script.push(`const emit = defineEmit(${text})`);
            }
            if (vueImports.length > 0) {
                script.unshift(`import { ${vueImports.join(', ')} } from 'vue'`);
            }
        }
        const sfc = [];
        if (script.length > 0) {
            sfc.push(getScriptStartTag(project.config.preferences.script), ...script, `</script>\n`);
        }
        sfc.push(`<template>${stringify(nodes, {
            indent: 2,
            initialIndent: 1,
            directive: project.config.preferences.template.directiveSyntax,
            replaceNodes,
        })}</template>\n`);
        // Generate text changes to import newly created component.
        const code = sfc.join('\n');
        const { name, changes, renameLocation } = getImportEditForComponent(document.container, info, componentFileName, project.config);
        // Generate text changes to use newly created component in <template> of current component.
        const templateReplacement = {
            newText: genTagForExtractedComponent({
                name,
                project,
                emits,
                models,
                identifiers,
                replaceNodes,
            }),
            span: {
                start: first(nodes).loc.start.offset,
                length: last(nodes).loc.end.offset - first(nodes).loc.start.offset,
            },
        };
        changes.push(templateReplacement);
        config.context.createVueDocument(componentFileName, code);
        return {
            renameFilename: document.container.fsPath,
            renameLocation: renameLocation,
            edits: [
                {
                    isNewFile: true,
                    fileName: componentFileName,
                    textChanges: [
                        {
                            span: {
                                start: 0,
                                length: 0,
                            },
                            newText: code,
                        },
                    ],
                },
                {
                    fileName: document.container.fsPath,
                    textChanges: changes,
                },
            ],
        };
    },
};
function genEmits(context, emits, models) {
    return `[${[...emits, ...models.map((id) => `onUpdate:${id}`)]
        .map((key) => `'${key}'`)
        .join(', ')}]`;
}
function genProps(context, identifiers) {
    return `[${identifiers.map((id) => `'${id}'`).join(', ')}]`;
}
function genTagForExtractedComponent({ name, project, emits, models, identifiers, replaceNodes, }) {
    const useShorthand = project.config.preferences.template.directiveSyntax === 'shorthand';
    const skipNodes = new Set();
    replaceNodes.forEach((value, key) => {
        if (value == null)
            skipNodes.add(key);
    });
    const skipNodesTemplate = Array.from(skipNodes)
        .map((node) => stringify(node, {
        indent: 2,
        initialIndent: 0,
        directive: project.config.preferences.template.directiveSyntax,
    }))
        .join(' ');
    const modelsTemplate = Array.from(models)
        .map((id) => `v-model:${id}="${id}"`)
        .join(' ');
    const vOnDirText = useShorthand ? '@' : 'v-on:';
    const vBindDirText = useShorthand ? ':' : 'v-bind:';
    const emitsTemplate = Array.from(emits.entries())
        .map(([key, value]) => `${vOnDirText}${key}="${value}"`)
        .join(' ');
    const bindsTemplate = identifiers
        .map((id) => `${vBindDirText}${id}="${id}"`)
        .join(' ');
    const openTag = [
        name,
        skipNodesTemplate,
        modelsTemplate,
        emitsTemplate,
        bindsTemplate,
    ]
        .filter(Boolean)
        .join(' ');
    return `<${openTag} />`;
}
function genEmitsOptionsText({ emits, models, }) {
    const text = [
        ...emits.keys(),
        ...Array.from(models).map((id) => `onUpdate:${id}`),
    ]
        .map((key) => `'${key}'`)
        .join(', ');
    return text !== '' ? `emits: [${text}],` : '';
}
function genPropsOptionsText({ identifiers, models, }) {
    const text = [...identifiers, ...models].map((id) => `'${id}'`).join(', ');
    return text !== '' ? `props: [${text}],` : '';
}
function genComponentsOptionsText({ components, info, addImport, fileName, componentFileName, }) {
    const text = info.components
        .map(({ name, source, aliases }) => {
        if (aliases.some((alias) => components.has(alias))) {
            addImport(genComponentImport(fileName, componentFileName, source));
            const registration = name === source.localName
                ? source.localName
                : `${name}: ${source.localName}`;
            return registration;
        }
    })
        .filter(Boolean)
        .join(', ');
    return text !== '' ? `components: {${text}},` : '';
}
function genComponentImport(oldFileName, newFileName, source) {
    let moduleName = source.moduleName;
    if (moduleName.startsWith('.')) {
        moduleName = Path.posix.relative(Path.posix.dirname(newFileName), Path.posix.resolve(Path.posix.dirname(oldFileName), moduleName));
        if (!moduleName.startsWith('.')) {
            moduleName = `./${moduleName}`;
        }
    }
    return `import ${source.exportName != null
        ? `{ ${source.localName === source.exportName
            ? ``
            : `${source.exportName} as `}${source.localName} }`
        : `${source.localName}`} from '${moduleName}'`;
}
function detectConditionType(parent, nodes) {
    const conditions = nodes.map((node) => {
        if (isElementNode(node)) {
            return node.props.find((prop) => isDirectiveNode(prop) && /^(if|else-if|else)$/.test(prop.name));
        }
    });
    const nonNullConditions = conditions.filter((condition) => condition != null);
    if (nonNullConditions.length > 0) {
        if (conditions.length === 1)
            return 'preserve';
        const firstCondition = first(conditions);
        const lastCondition = last(conditions);
        if ((firstCondition == null || firstCondition.name === 'if') &&
            (lastCondition == null || lastCondition.name === 'else')) {
            return 'extract';
        }
        if (firstCondition != null && firstCondition.name !== 'if')
            return 'partial';
        const nextSibling = findNextSibling(parent.children, last(nodes));
        const nextCondition = isElementNode(nextSibling)
            ? nextSibling.props.find((prop) => isDirectiveNode(prop) && /^(else-if|else)$/.test(prop.name))
            : undefined;
        if (nextCondition == null)
            return 'extract';
        return 'partial';
    }
    else {
        return 'none';
    }
}
function getScriptStartTag(preferences) {
    return preferences.mode === 'normal'
        ? preferences.language !== 'js'
            ? `<script lang=${JSON.stringify(preferences.language)}>`
            : `<script>`
        : preferences.language !== 'js'
            ? `<script lang=${JSON.stringify(preferences.language)} setup>`
            : `<script setup>`;
}
function findNodes(document, position) {
    return document.ast != null
        ? typeof position === 'number'
            ? findTemplateChildrenInRange(document.ast, position, position)
            : findTemplateChildrenInRange(document.ast, position.pos, position.end)
        : [];
}
function findParentNode(ast, nodes) {
    if (nodes.length === 0)
        return ast;
    const search = nodes[0];
    let result;
    traverseFast(ast, (node, _, stop) => {
        if (isElementNode(node)) {
            if (node.children.includes(search)) {
                result = node;
                stop();
            }
        }
    });
    return result !== null && result !== void 0 ? result : ast;
}
function getImportEditForComponent(document, info, fileName, config) {
    const name = getComponentName(fileName);
    const { script, scriptSetup } = document.descriptor;
    const relativeFileName = `./${Path.posix.relative(Path.posix.dirname(document.fsPath), fileName)}`;
    const importStatement = `import ${name} from '${relativeFileName}';\n`;
    let renameLocation = importStatement.indexOf(relativeFileName) + relativeFileName.length - 3;
    const changes = registerLocalComponentWithSource(document, info, { moduleName: relativeFileName, localName: name }, config.preferences.script, importStatement);
    if (scriptSetup != null) {
        renameLocation += scriptSetup.loc.start.offset;
    }
    else if (script != null) {
        renameLocation += script.loc.start.offset;
    }
    else {
        renameLocation += changes[0].newText.indexOf(importStatement);
    }
    return { name, changes, renameLocation };
}

function detectRefsAndProps(service, document, info, node) {
    const refs = [];
    const props = [];
    const program = service.getProgram();
    const sourceFile = program === null || program === void 0 ? void 0 : program.getSourceFile(getScriptFileName(document));
    if (program != null && sourceFile != null) {
        const checker = program.getTypeChecker();
        const findNodeAt = (start, end, container = sourceFile) => {
            const node = container
                .getChildren()
                .find((node) => node.pos <= start && end <= node.end);
            if (node != null) {
                if (node.pos === start && node.pos === end)
                    return node;
                else
                    return findNodeAt(start, end, node);
            }
            else {
                return container;
            }
        };
        node.scope.identifiers.forEach((id) => {
            if (node.scope.getBinding(id) == null && id in info.identifierSource) {
                const source = info.identifierSource[id];
                if (source.name === 'setup' || source.name.startsWith('scriptSetup:')) {
                    const node = findNodeAt(source.loc.start.offset, source.loc.end.offset);
                    if (node != null) {
                        const type = checker.getTypeAtLocation(node);
                        const str = checker.typeToString(type);
                        if (/^(Ref|ComputedRef)<.*?>$/.test(str.trim())) {
                            refs.push(id);
                        }
                    }
                }
                else if (source.name === 'props') {
                    props.push(id);
                }
            }
        });
    }
    return { refs, props };
}

function genSetupFnParams(info, node) {
    var _a, _b, _c, _d, _e, _f, _g;
    const changes = [];
    if (info.fnSetupOption != null) {
        const needsProps = node.scope.identifiers.some((id) => {
            var _a;
            if (node.scope.getBinding(id) == null) {
                return ((_a = info.identifierSource[id]) === null || _a === void 0 ? void 0 : _a.name) === 'props';
            }
            return false;
        }) && ((_a = info.fnSetupOption.props) === null || _a === void 0 ? void 0 : _a.rest) == null;
        const needsEmit = node.scope.identifiers.includes('$emit') &&
            ((_b = info.fnSetupOption.context) === null || _b === void 0 ? void 0 : _b.identifiers.emit) == null &&
            ((_c = info.fnSetupOption.context) === null || _c === void 0 ? void 0 : _c.rest) == null;
        const needsAttrs = node.scope.identifiers.includes('$attrs') &&
            ((_d = info.fnSetupOption.context) === null || _d === void 0 ? void 0 : _d.identifiers.attrs) == null &&
            ((_e = info.fnSetupOption.context) === null || _e === void 0 ? void 0 : _e.rest) == null;
        const needsSlots = node.scope.identifiers.includes('$slots') &&
            ((_f = info.fnSetupOption.context) === null || _f === void 0 ? void 0 : _f.identifiers.slots) == null &&
            ((_g = info.fnSetupOption.context) === null || _g === void 0 ? void 0 : _g.rest) == null;
        const needsContext = needsAttrs || needsEmit || needsSlots;
        if (needsProps || needsContext) {
            if (info.fnSetupOption.context != null) {
                const index = info.fnSetupOption.context.loc.source.indexOf('{');
                if (index >= 0) {
                    const offset = getPaddingLength(info.fnSetupOption.context.loc.source, index + 1) +
                        index +
                        1;
                    const args = [
                        needsSlots ? 'slots' : undefined,
                        needsAttrs ? 'attrs' : undefined,
                        needsEmit ? 'emit' : undefined,
                    ]
                        .filter(Boolean)
                        .join(', ');
                    changes.push({
                        newText: `${args}, `,
                        span: {
                            start: info.fnSetupOption.context.loc.start.offset + offset,
                            length: 0,
                        },
                    });
                }
            }
            else if (info.fnSetupOption.props != null) {
                const hasParen = info.fnSetupOption.loc.source.indexOf('(') <
                    info.fnSetupOption.props.loc.start.offset;
                const props = info.fnSetupOption.props.loc.source;
                changes.push({
                    newText: hasParen ? `${props}, context` : `(${props}, context)`,
                    span: {
                        start: info.fnSetupOption.props.loc.start.offset,
                        length: info.fnSetupOption.props.loc.source.length,
                    },
                });
            }
            else {
                changes.push({
                    newText: needsContext ? 'props, context' : 'props',
                    span: {
                        start: info.fnSetupOption.loc.start.offset +
                            info.fnSetupOption.loc.source.indexOf('(') +
                            1,
                        length: 0,
                    },
                });
            }
        }
    }
    return changes;
}

function getContextIdentifiers(info) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j;
    if (info.fnSetupOption != null) {
        const { context, props } = info.fnSetupOption;
        if ((context === null || context === void 0 ? void 0 : context.rest) != null) {
            return {
                props: (_a = props === null || props === void 0 ? void 0 : props.rest) !== null && _a !== void 0 ? _a : 'props',
                slots: (_b = context.identifiers.slots) !== null && _b !== void 0 ? _b : `${context.rest}.slots`,
                attrs: (_c = context.identifiers.attrs) !== null && _c !== void 0 ? _c : `${context.rest}.attrs`,
                emit: (_d = context.identifiers.emit) !== null && _d !== void 0 ? _d : `${context.rest}.emit`,
            };
        }
        else if (context != null) {
            return {
                props: (_e = props === null || props === void 0 ? void 0 : props.rest) !== null && _e !== void 0 ? _e : 'props',
                slots: (_f = context === null || context === void 0 ? void 0 : context.identifiers.slots) !== null && _f !== void 0 ? _f : `slots`,
                attrs: (_g = context === null || context === void 0 ? void 0 : context.identifiers.attrs) !== null && _g !== void 0 ? _g : `attrs`,
                emit: (_h = context === null || context === void 0 ? void 0 : context.identifiers.emit) !== null && _h !== void 0 ? _h : `emit`,
            };
        }
        else {
            return {
                props: (_j = props === null || props === void 0 ? void 0 : props.rest) !== null && _j !== void 0 ? _j : 'props',
                slots: `context.slots`,
                attrs: `context.attrs`,
                emit: `context.emit`,
            };
        }
    }
    return {
        props: 'props',
        slots: `context.slots`,
        attrs: `context.attrs`,
        emit: `context.emit`,
    };
}

const RefactorExtractComputed = {
    version: '*',
    name: REFACTORS.EXTRACT_COMPUTED,
    findRefactors({ helpers, service }, fileName, position, preferences) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        const extractAsComputedProperty = {
            name: encode({ target: 'computed' }),
            description: 'Extract to computed property',
        };
        const extractAsComputedValue = {
            name: encode({ target: 'setupComputed' }),
            description: 'Extract to computed value',
        };
        const refactor = {
            name: REFACTORS.EXTRACT_COMPUTED,
            description: 'Extract to computed',
            inlineable: true,
            actions: [extractAsComputedProperty, extractAsComputedValue],
        };
        const { node, ancestors, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (document != null && isSimpleExpressionNode(node)) {
            const parent = last(ancestors).node;
            if (document.container.descriptor.scriptSetup != null) {
                extractAsComputedProperty.notApplicableReason = (_a = extractAsComputedProperty.notApplicableReason) !== null && _a !== void 0 ? _a : 'Not supported when using <script setup>';
            }
            else {
                const info = helpers.getComponentInfo(document.container);
                node.scope.identifiers.forEach((id) => {
                    var _a, _b, _c;
                    const binding = node.scope.getBinding(id);
                    if (binding != null) {
                        extractAsComputedProperty.notApplicableReason = (_a = extractAsComputedProperty.notApplicableReason) !== null && _a !== void 0 ? _a : `Expression uses local variables`;
                        extractAsComputedValue.notApplicableReason = (_b = extractAsComputedValue.notApplicableReason) !== null && _b !== void 0 ? _b : `Expression uses local variables`;
                    }
                    else {
                        const source = info.identifierSource[id];
                        if (source != null) {
                            if (/^(methods|computed|data)$/.test(source.name)) {
                                extractAsComputedValue.notApplicableReason = (_c = extractAsComputedValue.notApplicableReason) !== null && _c !== void 0 ? _c : `Cannot use ${source.name} from options API in setup() function`;
                            }
                        }
                    }
                });
            }
            if (isSimpleIdentifier(node.content)) {
                extractAsComputedProperty.notApplicableReason = (_b = extractAsComputedProperty.notApplicableReason) !== null && _b !== void 0 ? _b : 'Already extracted';
                extractAsComputedValue.notApplicableReason = (_c = extractAsComputedValue.notApplicableReason) !== null && _c !== void 0 ? _c : 'Already extracted';
            }
            else if (isDirectiveNode(parent) && // Unsupported directives.
                (parent.name === 'slot' ||
                    parent.name === 'for' ||
                    parent.name === 'on' ||
                    parent.name === 'model')) {
                extractAsComputedProperty.notApplicableReason = (_d = extractAsComputedProperty.notApplicableReason) !== null && _d !== void 0 ? _d : `Not supported in v-${parent.name}`;
                extractAsComputedValue.notApplicableReason = (_e = extractAsComputedValue.notApplicableReason) !== null && _e !== void 0 ? _e : `Not supported in v-${parent.name}`;
            }
        }
        else {
            extractAsComputedProperty.notApplicableReason = (_f = extractAsComputedProperty.notApplicableReason) !== null && _f !== void 0 ? _f : 'Only expressions can be extracted as computed property';
            extractAsComputedValue.notApplicableReason = (_g = extractAsComputedValue.notApplicableReason) !== null && _g !== void 0 ? _g : 'Only expressions can be extracted as computed value';
        }
        if (extractAsComputedProperty.notApplicableReason == null &&
            ((_h = helpers.getVueDocument(fileName)) === null || _h === void 0 ? void 0 : _h.descriptor.scriptSetup) != null) {
            extractAsComputedProperty.notApplicableReason =
                'No supported when using <script setup>';
        }
        if (typeof position === 'object' &&
            position.pos !== position.end &&
            (extractAsComputedProperty.notApplicableReason == null ||
                extractAsComputedValue.notApplicableReason == null)) ;
        return [refactor];
    },
    applyRefactor({ helpers, context, service }, fileName, _, position, refactorName, action) {
        if (refactorName !== REFACTORS.EXTRACT_COMPUTED)
            return;
        const args = decode(action);
        const document = helpers.getVueDocument(fileName);
        const { node } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node) && document !== null) {
            const project = context.getVueProjectForFile(document.fsPath, true);
            const shouldRewriteContext = args.target === 'computed' && document.descriptor.scriptSetup == null;
            let start = node.loc.start.offset;
            let end = node.loc.end.offset;
            if (typeof position !== 'number' && position.pos !== position.end) {
                start = position.pos;
                end = position.end;
            }
            const content = node.content.substring(start - node.loc.start.offset, end - node.loc.start.offset);
            const info = helpers.getComponentInfo(document);
            const code = transformToFunction(content, {
                name: 'anonymous',
                kind: 'expression',
                args: [],
                rewrite: shouldRewriteContext
                    ? { context: 'this' }
                    : {
                        ...detectRefsAndProps(service, document, info, node),
                        identifiers: getContextIdentifiers(info),
                    },
            });
            const name = 'newComputed';
            const replaceExp = {
                newText: name,
                span: {
                    start: start,
                    length: end - start,
                },
            };
            const { changes } = registerComponentAPI(document, info, args.target, name, code, project.config.preferences.script);
            if (info.fnSetupOption != null) {
                changes.push(...genSetupFnParams(info, node));
            }
            return {
                renameFilename: document.fsPath,
                renameLocation: start,
                edits: [
                    {
                        fileName: document.fsPath,
                        textChanges: [...changes, replaceExp],
                    },
                ],
            };
        }
    },
};

const RefactorExtractMethod = {
    version: '*',
    name: REFACTORS.EXTRACT_METHOD,
    findRefactors({ helpers }, fileName, position) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        const extractAsMethod = {
            name: encode({ target: 'methods' }),
            description: 'Extract to method',
        };
        const extractAsFunction = {
            name: encode({ target: 'setupMethods' }),
            description: 'Extract to function',
        };
        const refactor = {
            name: REFACTORS.EXTRACT_METHOD,
            description: 'Extract to function/method',
            inlineable: true,
            actions: [extractAsMethod, extractAsFunction],
        };
        const { node, ancestors, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (document != null && isSimpleExpressionNode(node)) {
            const parent = last(ancestors).node;
            if (document.container.descriptor.scriptSetup != null) {
                extractAsMethod.notApplicableReason = (_a = extractAsMethod.notApplicableReason) !== null && _a !== void 0 ? _a : 'Not supported when using <script setup>';
            }
            else {
                const info = helpers.getComponentInfo(document.container);
                node.scope.identifiers.forEach((id) => {
                    var _a;
                    const binding = node.scope.getBinding(id);
                    const source = info.identifierSource[id];
                    if (source != null && binding == null) {
                        if (/^(methods|computed|data)$/.test(source.name)) {
                            extractAsFunction.notApplicableReason = (_a = extractAsFunction.notApplicableReason) !== null && _a !== void 0 ? _a : `Cannot use ${source.name} from options API in setup() function`;
                        }
                    }
                });
            }
            if (isSimpleIdentifier(node.content)) {
                extractAsMethod.notApplicableReason = (_b = extractAsMethod.notApplicableReason) !== null && _b !== void 0 ? _b : 'Already extracted';
                extractAsFunction.notApplicableReason = (_c = extractAsFunction.notApplicableReason) !== null && _c !== void 0 ? _c : 'Already extracted';
            }
            else if (isDirectiveNode(parent) &&
                (parent.name === 'slot' || parent.name === 'for')) {
                extractAsMethod.notApplicableReason = (_d = extractAsMethod.notApplicableReason) !== null && _d !== void 0 ? _d : `Not supported in v-${parent.name}`;
                extractAsFunction.notApplicableReason = (_e = extractAsFunction.notApplicableReason) !== null && _e !== void 0 ? _e : `Not supported in v-${parent.name}`;
            }
        }
        else {
            extractAsMethod.notApplicableReason = (_f = extractAsMethod.notApplicableReason) !== null && _f !== void 0 ? _f : 'Only expressions can be extracted as method';
            extractAsFunction.notApplicableReason = (_g = extractAsFunction.notApplicableReason) !== null && _g !== void 0 ? _g : 'Only expressions can be extracted as function';
        }
        if (extractAsMethod.notApplicableReason == null &&
            ((_h = helpers.getVueDocument(fileName)) === null || _h === void 0 ? void 0 : _h.descriptor.scriptSetup) != null) {
            extractAsMethod.notApplicableReason =
                'No supported when using <script setup>';
        }
        return [refactor];
    },
    applyRefactor({ helpers, context, service }, fileName, _, position, refactorName, action) {
        var _a;
        if (refactorName !== REFACTORS.EXTRACT_METHOD)
            return;
        const args = decode(action);
        const document = helpers.getVueDocument(fileName);
        const { node, ancestors } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node) && document !== null) {
            const project = context.getVueProjectForFile(document.fsPath, true);
            const parent = (_a = last(ancestors)) === null || _a === void 0 ? void 0 : _a.node;
            const isEventHandler = isDirectiveNode(parent) && parent.name === 'on' && parent.exp === node;
            const shouldRewriteContext = args.target === 'methods' && document.descriptor.scriptSetup == null;
            const methodArgs = [];
            let shouldCallEventHandler = false;
            if (isEventHandler &&
                /(?<![a-z.[])\$event(?=[^a-zA-Z0-9_$])/.test(node.content)) {
                methodArgs.push('$event');
            }
            node.scope.identifiers.forEach((id) => {
                if (node.scope.getBinding(id) != null) {
                    methodArgs.push(id);
                    shouldCallEventHandler = true;
                }
            });
            const info = helpers.getComponentInfo(document);
            const code = transformToFunction(node.content, {
                name: 'anonymous',
                kind: isEventHandler ? 'statement' : 'expression',
                args: methodArgs,
                rewrite: shouldRewriteContext
                    ? { context: 'this' }
                    : {
                        ...detectRefsAndProps(service, document, info, node),
                        identifiers: getContextIdentifiers(info),
                    },
            });
            const name = args.target === 'methods' ? 'newMethod' : 'newFunction';
            const replaceExp = {
                newText: isEventHandler && !shouldCallEventHandler
                    ? name
                    : `${name}(${methodArgs.join(', ')})`,
                span: {
                    start: node.loc.start.offset,
                    length: node.loc.end.offset - node.loc.start.offset,
                },
            };
            const { changes } = registerComponentAPI(document, info, args.target, name, code, project.config.preferences.script);
            if (info.fnSetupOption != null) {
                changes.push(...genSetupFnParams(info, node));
            }
            return {
                renameFilename: document.fsPath,
                renameLocation: node.loc.start.offset,
                edits: [
                    {
                        fileName: document.fsPath,
                        textChanges: [...changes, replaceExp],
                    },
                ],
            };
        }
    },
};

const REFACTOR_PROVIDERS = [
    RefactorExtractComponent,
    RefactorExtractMethod,
    RefactorExtractComputed,
];
const SCRIPT_REFACTOR_PROVIDERS = [AddToSetupContextRefactor];
/*
☑︎  1. Extract to component
☑︎  2. Extract to computed property (full and partial expression)
☑︎  3. Extract to method (v-on and other expression)
☑︎  4. Add to return statement in setup() function
◼︎  5. Move v-if inside component
☐  6. Convert options API to functional API
☒  7. Convert <script> to <script setup>
☒  8. Convert <script setup> to <script>
☒  9. Convert to ref syntax sugar
☒ 10. Remove ref syntax sugar
☐ 11. Convert template to render function
 */

const RenameComponentTag = {
    version: '*',
    name: 'component-tag',
    canRename(config, fileName, position, options) {
        const { node, document } = config.helpers.findTemplateNodeAtPosition(fileName, position);
        if (isComponentNode(node) &&
            isPositionInTagName(position, node) &&
            document != null) {
            const info = config.helpers.getComponentInfo(document.container);
            const name = node.tag;
            const component = info.components.find((component) => component.name === name);
            if (component == null) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Cannot rename global component',
                };
            }
            return {
                canRename: true,
                displayName: node.tag,
                fullDisplayName: node.tag,
                kind: config.context.typescript.ScriptElementKind.unknown,
                kindModifiers: 'componentTagName',
                fileToRename: undefined,
                triggerSpan: {
                    start: node.loc.start.offset + 1,
                    length: node.tag.length,
                },
            };
        }
    },
    applyRename({ helpers, service }, fileName, position, findInStrings, findInComments) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (document == null)
            return;
        if (!isComponentNode(node))
            return;
        const info = helpers.getComponentInfo(document.container);
        const name = node.tag;
        const component = info.components.find((component) => component.name === name);
        if (component == null)
            return;
        const scriptFileName = document.container.descriptor.scriptSetup != null
            ? document.container.getDocumentFileName(SCRIPT_SETUP_BLOCK_SELECTOR)
            : document.container.getDocumentFileName(SCRIPT_BLOCK_SELECTOR);
        const renameLocations = [];
        // if component is registered with an alias, then update to reflect in template.
        if (component.source.localName !== component.name) {
            renameLocations.push({
                fileName: document.container.fsPath,
                textSpan: {
                    start: component.loc.start.offset,
                    length: component.loc.source.length,
                },
                ...computeIdentifierReplacement(component.loc.source, component.name),
            });
        }
        // update the import statement.
        else {
            const { prefixText } = computeIdentifierReplacement(component.loc.source, component.source.localName);
            const locations = service.findRenameLocations(scriptFileName, component.loc.start.offset + prefixText.length, findInStrings, findInComments);
            if (locations != null) {
                renameLocations.push(...locations.filter((location) => location.fileName === scriptFileName));
            }
        }
        if (document.ast != null) {
            getChangesForComponentTagRename(document.ast, component).forEach((textSpan) => renameLocations.push({
                fileName: document.container.fsPath,
                textSpan,
            }));
        }
        return renameLocations;
    },
    applyFileRename(config, containerFile, oldFileName, newFileName, options, preferences) {
        const fileTextChanges = [];
        if (isVueFile(oldFileName)) {
            const document = config.helpers.getRenderDoc(containerFile);
            if (document != null) {
                const name = getComponentName(oldFileName);
                const newName = getComponentName(newFileName);
                const info = config.helpers.getComponentInfo(document.container);
                const component = info.components.find((component) => component.name === name);
                if (component != null) {
                    const { scriptSetup, script } = document.container.descriptor;
                    const scriptFileName = scriptSetup != null
                        ? document.container.getDocumentFileName('scriptSetup')
                        : script != null
                            ? document.container.getDocumentFileName('script')
                            : null;
                    if (scriptFileName != null) {
                        const position = Math.max(component.loc.end.offset - 2, component.loc.start.offset);
                        const renameInfo = config.service.getRenameInfo(scriptFileName, position);
                        if (renameInfo.canRename) {
                            const edits = config.service.findRenameLocations(scriptFileName, position, false, false);
                            if (edits != null) {
                                const fileName = document.container.fsPath;
                                fileTextChanges.push({
                                    fileName: fileName,
                                    textChanges: edits
                                        .filter((edit) => edit.fileName === scriptFileName)
                                        .map((edit) => ({ span: edit.textSpan, newText: newName })),
                                });
                            }
                        }
                        // update tags in template
                        if (document.ast != null &&
                            name === component.source.localName &&
                            component.name === component.source.localName) {
                            const textChanges = [];
                            if (component.loc.source.trim() !== name) {
                                // Rename alias in components options
                                textChanges.push({
                                    span: {
                                        start: component.loc.start.offset,
                                        length: component.loc.source.indexOf(':'),
                                    },
                                    newText: newName,
                                });
                            }
                            traverseFast(document.ast, (node) => {
                                if (isComponentNode(node)) {
                                    if (component.aliases.includes(node.tag)) {
                                        textChanges.push({
                                            newText: newName,
                                            span: {
                                                start: node.loc.start.offset + 1,
                                                length: node.tag.length,
                                            },
                                        });
                                        if (!node.isSelfClosing) {
                                            textChanges.push({
                                                newText: newName,
                                                span: {
                                                    start: node.loc.start.offset +
                                                        node.loc.source.lastIndexOf('</' + node.tag) +
                                                        2,
                                                    length: node.tag.length,
                                                },
                                            });
                                        }
                                    }
                                }
                            });
                            fileTextChanges.push({
                                fileName: document.container.fsPath,
                                textChanges: textChanges,
                            });
                        }
                    }
                }
            }
        }
        if (fileTextChanges.length > 0)
            return fileTextChanges;
    },
};
function getChangesForComponentTagRename(ast, component) {
    const locations = [];
    traverseFast(ast, (node) => {
        if (isComponentNode(node)) {
            if (component.aliases.includes(node.tag)) {
                locations.push({
                    start: node.loc.start.offset + 1,
                    length: node.tag.length,
                });
                if (!node.isSelfClosing) {
                    locations.push({
                        start: node.loc.start.offset +
                            node.loc.source.lastIndexOf('</' + node.tag) +
                            2,
                        length: node.tag.length,
                    });
                }
            }
        }
    });
    return locations;
}
function isPositionInTagName(position, node) {
    return (
    // In start tag.
    position <= node.loc.start.offset + node.tag.length ||
        // In end tag.
        (!node.isSelfClosing &&
            node.loc.start.offset + node.loc.source.lastIndexOf('</') <= position));
}

function defineCompletionProvider(options) {
    return options;
}

const TemplateContextCompletionProvider = defineCompletionProvider({
    version: '*',
    getCompletionsAtPosition({ helpers, service, context }, fileName, position, options) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document) &&
            isSimpleExpressionNode(node) &&
            (options === null || options === void 0 ? void 0 : options.triggerCharacter) !== '.' // Not a member expression
        ) {
            context.log('@@DEBUG TemplateContext Completion');
            const result = service.getCompletionsAtPosition(document.fsPath, document.contextCompletionsTriggerOffset, {
                ...options,
                triggerCharacter: '.',
                includeAutomaticOptionalChainCompletions: true,
                includeCompletionsForModuleExports: false,
                includeCompletionsWithInsertText: true,
                includePackageJsonAutoImports: 'off',
                provideRefactorNotApplicableReason: false,
                disableSuggestions: true,
            });
            if (isNotNull(result)) {
                result.entries.filter((entry) => {
                    if (entry.name.startsWith('$')) {
                        entry.sortText = '9';
                    }
                    return entry.name.startsWith('_');
                });
            }
            return result;
        }
        return undefined;
    },
    getCompletionEntryDetails({ helpers, service }, fileName, position, entryName, formatOptions, source, preferences) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document) && isSimpleExpressionNode(node)) {
            return service.getCompletionEntryDetails(document.fsPath, document.contextCompletionsTriggerOffset, entryName, formatOptions, source, preferences);
        }
        return undefined;
    },
    getCompletionEntrySymbol({ helpers, service }, fileName, position, name, source) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document) && isSimpleExpressionNode(node)) {
            return service.getCompletionEntrySymbol(document.fsPath, document.contextCompletionsTriggerOffset, name, source);
        }
        return undefined;
    },
});

const disallowedIdentifiers = new Set([
    'arguments',
    'globalThis',
    'default',
    '_ctx',
]);
const allowedGlobals = new Set([
    'Infinity',
    'undefined',
    'NaN',
    'isFinite',
    'isNaN',
    'parseFloat',
    'parseInt',
    'decodeURI',
    'decodeURIComponent',
    'encodeURI',
    'encodeURIComponent',
    'Math',
    'Number',
    'Date',
    'Array',
    'Object',
    'Boolean',
    'String',
    'RegExp',
    'Map',
    'Set',
    'JSON',
    'Intl',
]);
const allowedKeywords = new Set([
    'function',
    'return',
    'true',
    'false',
    'null',
    'typeof',
    'instanceof',
    'async',
    'await',
    'new',
]);
const TemplateExpressionCompletionProvider = defineCompletionProvider({
    version: '*',
    getCompletionsAtPosition({ helpers, service, context }, fileName, position, options) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document) && isSimpleExpressionNode(node)) {
            const loc = document.tryGetGeneratedOffset(position);
            if (isNotNull(loc)) {
                context.log('@@DEBUG TemplateExpression Completion');
                const result = service.getCompletionsAtPosition(document.fsPath, loc, options);
                if (isNotNull(result)) {
                    const isPropertyCompletion = (options === null || options === void 0 ? void 0 : options.triggerCharacter) === '.';
                    result.entries = result.entries.filter((entry) => {
                        if (isNotNull(entry.source))
                            return false; // Ignore externals.
                        if (disallowedIdentifiers.has(entry.name))
                            return false;
                        if (entry.kind === 'keyword') {
                            return allowedKeywords.has(entry.name);
                        }
                        if (isPropertyCompletion)
                            return true;
                        if (entry.kind === 'var' ||
                            entry.kind === 'module' ||
                            entry.kind === 'function') {
                            return allowedGlobals.has(entry.name);
                        }
                        return false;
                    });
                }
                return result;
            }
        }
        return undefined;
    },
    getCompletionEntryDetails({ helpers, service }, fileName, position, entryName, formatOptions, source, preferences) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document) && isSimpleExpressionNode(node)) {
            const loc = document.tryGetGeneratedOffset(position);
            if (isNotNull(loc)) {
                return service.getCompletionEntryDetails(document.fsPath, loc, entryName, formatOptions, source, preferences);
            }
        }
        return undefined;
    },
    getCompletionEntrySymbol({ helpers, service }, fileName, position, name, source) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document) && isSimpleExpressionNode(node)) {
            const loc = document.tryGetGeneratedOffset(position);
            if (isNotNull(loc)) {
                return service.getCompletionEntrySymbol(document.fsPath, loc, name, source);
            }
        }
        return undefined;
    },
});

const BUILTIN_DIRECTIVES = [];
const BUILTIN_DIRECTIVES_DETAIL = {
    'v-show': {
        name: 'v-show',
        kind: 'JSX attribute',
        kindModifiers: 'directive',
        displayParts: [],
        documentation: [
            {
                text: `[See documentation](https://vuejs.org)`,
                kind: 'markdown',
            },
        ],
        tags: [{ name: 'version', text: '3.0.1' }],
    },
};

const TemplatePropCompletionProvider = defineCompletionProvider({
    version: '*',
    getCompletionsAtPosition(config, fileName, position, options) {
        const completion = getCompletionType(config, fileName, position, options === null || options === void 0 ? void 0 : options.triggerCharacter);
        if (completion.type === 'none')
            return;
        const result = {
            isGlobalCompletion: false,
            isMemberCompletion: false,
            isNewIdentifierLocation: false,
            entries: [],
        };
        const { document, jsxPosition, element } = completion;
        if (element.tag === 'slot') ;
        else if (completion.type === 'directive') {
            if (completion.directive != null) ;
        }
        else {
            // TODO: Provide v-if etc.
            result.entries.push(...BUILTIN_DIRECTIVES);
            const jsxProps = config.service.getCompletionsAtPosition(document.fsPath, jsxPosition, {
                triggerCharacter: options === null || options === void 0 ? void 0 : options.triggerCharacter,
            });
            if (isArray(jsxProps === null || jsxProps === void 0 ? void 0 : jsxProps.entries)) {
                jsxProps === null || jsxProps === void 0 ? void 0 : jsxProps.entries.forEach((entry) => {
                    result.entries.push(entry); // TODO: Process to look like attribute.
                });
            }
        }
        return result;
    },
    getCompletionEntryDetails(config, fileName, position, entryName, formatOptions, source, preferences) {
        const completion = getCompletionType(config, fileName, position);
        if (completion.type === 'none')
            return;
        if (entryName in BUILTIN_DIRECTIVES_DETAIL) {
            return BUILTIN_DIRECTIVES_DETAIL[entryName];
        }
        return undefined;
    },
    getCompletionEntrySymbol(config, fileName, position, name, source) {
        const completion = getCompletionType(config, fileName, position);
        if (completion.type === 'none')
            return;
        return undefined;
    },
});
function getCompletionType({ helpers }, fileName, position, triggerCharacter) {
    const { node, document, ancestors } = helpers.findTemplateNodeAtPosition(fileName, position);
    if (isNotNull(document)) {
        if (isElementNode(node)) {
            // Is position after tag name?
            if (position >= node.loc.start.offset + node.tag.length + 2) {
                const jsxPosition = document.tryGetGeneratedOffset(node.loc.start.offset + 1);
                // Should be non-null unless there's is a unrecoverable syntax error in template.
                if (jsxPosition != null) {
                    return {
                        type: 'attribute',
                        document: document,
                        element: node,
                        jsxPosition: jsxPosition + pascalCase(node.tag).length + 1,
                    };
                }
            }
        }
        else if (isAttributeNode(node)) {
            const element = last(ancestors).node;
            const jsxPosition = document.tryGetGeneratedOffset(node.loc.start.offset);
            // This position is at start of the attribute.
            // TODO: Check if it needs to be the cursor position.
            if (jsxPosition != null) {
                return {
                    type: 'attribute',
                    document: document,
                    element: element,
                    attribute: node,
                    jsxPosition: jsxPosition,
                };
            }
        }
        else if (isDirectiveNode(node)) {
            const element = last(ancestors).node;
            const jsxPosition = document.tryGetGeneratedOffset(node.loc.start.offset);
            // This position is at start of the directive.
            // TODO: Check if it needs to be the cursor position.
            if (jsxPosition != null) {
                return {
                    type: 'directive',
                    document: document,
                    element: element,
                    directive: node,
                    jsxPosition: jsxPosition,
                };
            }
        }
    }
    return { type: 'none' };
}

// https://developer.mozilla.org/en-US/docs/Web/HTML/Element
const HTML_TAGS = new Set(('html,body,base,head,link,meta,style,title,address,article,aside,footer,' +
    'header,h1,h2,h3,h4,h5,h6,hgroup,nav,section,div,dd,dl,dt,figcaption,' +
    'figure,picture,hr,img,li,main,ol,p,pre,ul,a,b,abbr,bdi,bdo,br,cite,code,' +
    'data,dfn,em,i,kbd,keygen,mark,noindex,q,rp,rt,rtc,ruby,s,samp,small,span,strong,sub,sup,' +
    'time,u,var,wbr,area,audio,map,track,video,embed,object,param,source,' +
    'canvas,script,noscript,del,ins,caption,col,colgroup,table,thead,tbody,td,' +
    'th,tr,button,datalist,fieldset,form,input,label,legend,meter,optgroup,' +
    'option,output,progress,select,textarea,details,dialog,menu,' +
    'summary,template,blockquote,iframe,tfoot,webview').split(','));
const isHTMLTag = (tagName) => HTML_TAGS.has(tagName);
// https://developer.mozilla.org/en-US/docs/Web/SVG/Element
const SVG_TAGS = new Set(('svg,animate,animateMotion,animateTransform,circle,clipPath,color-profile,' +
    'defs,desc,discard,ellipse,feBlend,feColorMatrix,feComponentTransfer,' +
    'feComposite,feConvolveMatrix,feDiffuseLighting,feDisplacementMap,' +
    'feDistanceLight,feDropShadow,feFlood,feFuncA,feFuncB,feFuncG,feFuncR,' +
    'feGaussianBlur,feImage,feMerge,feMergeNode,feMorphology,feOffset,feDistantLight,' +
    'fePointLight,feSpecularLighting,feSpotLight,feTile,feTurbulence,filter,' +
    'foreignObject,g,hatch,hatchpath,image,line,linearGradient,marker,mask,' +
    'mesh,meshgradient,meshpatch,meshrow,metadata,mpath,path,pattern,' +
    'polygon,polyline,radialGradient,rect,set,solidcolor,stop,switch,symbol,' +
    'text,textPath,title,tspan,unknown,use,view').split(','));
const isSVGTag = (tagName) => SVG_TAGS.has(tagName);
const TemplateTagCompletionProvider = defineCompletionProvider({
    version: '*',
    getCompletionsAtPosition({ helpers, service, context }, fileName, position, options) {
        const { node, document, ancestors } = helpers.findTemplateNodeAtPosition(fileName, position);
        let result;
        if (isNotNull(document)) {
            if (isElementNode(node)) {
                const maxOffset = node.loc.start.offset + pascalCase(node.tag).length + 1;
                if (position <= maxOffset) {
                    const loc = document.tryGetGeneratedOffset(position);
                    if (isNotNull(loc)) {
                        result = service.getCompletionsAtPosition(document.fsPath, loc, {
                            triggerCharacter: '<',
                            ...options,
                        });
                    }
                }
            }
            else if (isTextNode(node) && /^(<|<>)$/.test(node.content.trim())) {
                result = service.getCompletionsAtPosition(document.fsPath, document.contextCompletionsTriggerOffset, { triggerCharacter: '<', ...options });
            }
            if (isNotNull(result)) {
                const { ScriptElementKind } = context.typescript;
                const needHTML = helpers.isFeatureEnabled('tagCompletions', 'html');
                const needSVG = helpers.isFeatureEnabled('tagCompletions', 'svg') &&
                    (ancestors.some((element) => isElementNode(element.node) && isSVGTag(element.node.tag)) ||
                        ancestors.length === 1);
                result.entries = result.entries.filter((entry) => {
                    if (isNotNull(entry.source)) {
                        if (isVirtualFile(entry.source)) {
                            if (isVirtualFileOfType(entry.source, MODULE_SELECTOR)) {
                                entry.kind = ScriptElementKind.jsxAttribute;
                            }
                        }
                    }
                    if (entry.kind === ScriptElementKind.jsxAttribute) {
                        if (!needHTML && isHTMLTag(entry.name))
                            return false;
                        if (!needSVG && isSVGTag(entry.name))
                            return false;
                        return true;
                    }
                    return false;
                });
                result.entries.forEach((entry) => {
                    if (isHTMLTag(entry.name))
                        entry.sortText = '8';
                    else if (isSVGTag(entry.name))
                        entry.sortText = '9';
                });
                const { typescript } = context;
                const knownTags = new Set(result.entries.map((entry) => entry.name));
                const info = helpers.getComponentInfo(document.container);
                info.components.forEach((component) => {
                    if (!knownTags.has(component.name)) {
                        component.aliases.forEach((alias) => {
                            knownTags.add(alias);
                            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                            result.entries.push({
                                name: alias,
                                kind: typescript.ScriptElementKind.jsxAttribute,
                                sortText: '0',
                            });
                        });
                    }
                });
                const project = context.getVueProjectForFile(document.container.fsPath, true);
                project.globalComponents.forEach((component) => {
                    if (!knownTags.has(component.name)) {
                        component.aliases.forEach((alias) => {
                            knownTags.add(alias);
                            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                            result.entries.push({
                                name: alias,
                                kind: typescript.ScriptElementKind.jsxAttribute,
                                sortText: '0',
                            });
                        });
                    }
                });
            }
        }
        return result;
    },
    getCompletionEntryDetails({ helpers, service, context }, fileName, position, entryName, formatOptions, source, preferences) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document)) {
            if (source == null) {
                const info = helpers.getComponentInfo(document.container);
                const component = info.components.find((component) => component.aliases.includes(entryName));
                if (component != null) {
                    // TODO: Get QuickInfo for component.
                    return {
                        name: entryName,
                        kind: context.typescript.ScriptElementKind.jsxAttribute,
                        kindModifiers: 'local component',
                        displayParts: [
                            {
                                text: component.source.exportName != null
                                    ? `import { ${component.source.exportName} as ${component.source.localName} } from '${component.source.moduleName}'`
                                    : `import ${component.source.localName} from '${component.source.moduleName}'`,
                                kind: 'ts',
                            },
                        ],
                    };
                }
                const project = context.getVueProjectForFile(document.container.fsPath, true);
                const globalComponent = project.globalComponents.find((component) => component.aliases.includes(entryName));
                if (globalComponent != null) {
                    // TODO: Get QuickInfo for globalComponent.
                    return {
                        name: entryName,
                        kind: context.typescript.ScriptElementKind.jsxAttribute,
                        kindModifiers: 'global component',
                        displayParts: [
                            {
                                text: globalComponent.source.exportName != null
                                    ? `import { ${globalComponent.source.exportName} as ${globalComponent.source.localName} } from '${globalComponent.source.moduleName}'`
                                    : `import ${globalComponent.source.localName} from '${globalComponent.source.moduleName}'`,
                                kind: 'ts',
                            },
                        ],
                    };
                }
            }
            if (isElementNode(node)) {
                const maxOffset = node.loc.start.offset + pascalCase(node.tag).length + 1;
                if (position <= maxOffset) {
                    const loc = document.tryGetGeneratedOffset(position);
                    if (isNotNull(loc)) {
                        const result = service.getCompletionEntryDetails(document.fsPath, loc, entryName, formatOptions, source, preferences);
                        if (result != null)
                            return result;
                    }
                }
            }
            else if (isTextNode(node) && /^(<|<>)$/.test(node.content.trim())) {
                const result = service.getCompletionEntryDetails(document.fsPath, document.tagCompletionsTriggerOffset, entryName, formatOptions, source, preferences);
                if (result != null)
                    return result;
            }
        }
        return undefined;
    },
    getCompletionEntrySymbol({ helpers, service }, fileName, position, name, source) {
        const { node, document } = helpers.findTemplateNodeAtPosition(fileName, position);
        if (isNotNull(document) && isElementNode(node)) {
            const maxOffset = node.loc.start.offset + pascalCase(node.tag).length + 1;
            if (position <= maxOffset) {
                const loc = document.tryGetGeneratedOffset(position);
                if (isNotNull(loc)) {
                    return service.getCompletionEntrySymbol(document.fsPath, document.contextCompletionsTriggerOffset, name, source);
                }
            }
        }
        return undefined;
    },
});

const TEMPLATE_COMPLETION_PROVIDERS = [
    TemplateTagCompletionProvider,
    TemplatePropCompletionProvider,
    TemplateContextCompletionProvider,
    TemplateExpressionCompletionProvider,
];

const GotoExternalComponentProp = {
    version: '*',
    getDefinitionAndBoundSpan({ helpers: h, context }, fileName, position) {
        var _a, _b, _c;
        const document = h.getRenderDoc(fileName);
        if ((document === null || document === void 0 ? void 0 : document.ast) == null)
            return;
        const targetNode = h.findTemplateNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(targetNode.node)) {
            const last = targetNode.ancestors.pop();
            if (last != null) {
                targetNode.node = last.node;
            }
        }
        let name;
        let node;
        if (isDirectiveNode(targetNode === null || targetNode === void 0 ? void 0 : targetNode.node) &&
            targetNode.node.name === 'bind' &&
            isSimpleExpressionNode(targetNode.node.arg)) {
            name = targetNode.node.arg.content;
            node = targetNode.node.arg;
        }
        else if (isAttributeNode(targetNode === null || targetNode === void 0 ? void 0 : targetNode.node)) {
            name = targetNode.node.name;
            node = targetNode.node;
        }
        else {
            return;
        }
        const parent = (_a = targetNode.ancestors[targetNode.ancestors.length - 1]) === null || _a === void 0 ? void 0 : _a.node;
        if (!isComponentNode(parent))
            return;
        const componentInfo = h.getComponentInfo(document.container);
        const component = componentInfo.components.find((component) => component.aliases.includes(parent.tag));
        if (component == null)
            return;
        const scriptFileName = ((_b = document.container.getDocument('script')) !== null && _b !== void 0 ? _b : document.container.getDocument('scriptSetup')).fsPath;
        const resolvedModule = h.getResolvedModule(scriptFileName, component.source.moduleName);
        if (resolvedModule == null)
            return;
        if (!isVirtualFile(resolvedModule.resolvedFileName))
            return;
        const externalComponentFilename = getContainingFile(resolvedModule.resolvedFileName);
        const componentDoc = h.getVueDocument(externalComponentFilename);
        if (componentDoc == null)
            return;
        const externalComponentInfo = h.getComponentInfo(componentDoc);
        const prop = externalComponentInfo.props.find((prop) => prop.name === name);
        if (prop == null)
            return;
        return {
            textSpan: {
                start: node.loc.start.offset,
                length: name.length,
            },
            definitions: [
                {
                    name: prop.name,
                    kind: context.typescript.ScriptElementKind.jsxAttribute,
                    fileName: componentDoc.fsPath,
                    textSpan: {
                        start: prop.loc.start.offset,
                        length: prop.loc.source.length,
                    },
                    containerKind: context.typescript.ScriptElementKind.unknown,
                    containerName: '',
                    contextSpan: ((_c = externalComponentInfo.options) === null || _c === void 0 ? void 0 : _c.properties.props) != null
                        ? {
                            start: externalComponentInfo.options.properties.props.loc.start
                                .offset,
                            length: externalComponentInfo.options.properties.props.loc.source
                                .length,
                        }
                        : undefined,
                },
            ],
        };
    },
};

const GotoScriptDefinitions = {
    version: '*',
    getDefinitionAndBoundSpan({ helpers: h, context, service }, fileName, position) {
        var _a;
        const document = h.getRenderDoc(fileName);
        if (document == null)
            return;
        const location = document.getGeneratedOffsetAt(position);
        if (location == null)
            return;
        const result = service.getDefinitionAndBoundSpan(fileName, location.offset);
        if (result == null)
            return;
        let isTextSpanSet = false;
        const definitions = [];
        const nodeAtCursor = h.findTemplateNodeAtPosition(document.fsPath, position);
        (_a = result.definitions) === null || _a === void 0 ? void 0 : _a.forEach((definition) => {
            var _a;
            if (h.isRenderFunctionFileName(definition.fileName)) {
                if (definition.fileName !== document.fsPath) {
                    context.log(`Unexpected bound span in ${document.fsPath} resolved to ${definition.fileName}`);
                    return;
                }
                if (!document.isInGeneratedRange(definition.textSpan.start)) {
                    const newResult = service.getDefinitionAndBoundSpan(definition.fileName, definition.textSpan.start);
                    if (newResult != null &&
                        Array.isArray(newResult.definitions) &&
                        newResult.definitions.length > 0) {
                        // TODO: Resolved contextSpan if `newResult` is resolved as return from setup()
                        definitions.push(...newResult.definitions);
                    }
                    else if (definition.kind === 'var') {
                        const componentInfo = h.getComponentInfo(document.container);
                        const prop = componentInfo.props.find((prop) => prop.name === definition.name);
                        if (prop != null) {
                            definitions.push({
                                ...definition,
                                fileName: document.container.fsPath,
                                textSpan: {
                                    start: prop.loc.start.offset,
                                    length: prop.loc.source.length,
                                },
                                containerKind: context.typescript.ScriptElementKind.unknown,
                                containerName: '',
                                contextSpan: ((_a = componentInfo.options) === null || _a === void 0 ? void 0 : _a.properties.props) != null
                                    ? {
                                        start: componentInfo.options.properties.props.loc.start
                                            .offset,
                                        length: componentInfo.options.properties.props.loc.source
                                            .length,
                                    }
                                    : undefined,
                            });
                        }
                    }
                }
                else {
                    const textSpan = h.getTextSpan(document, definition.textSpan, nodeAtCursor.node);
                    if (textSpan !== definition.textSpan) {
                        isTextSpanSet = true;
                        definition.textSpan = textSpan;
                        if (definition.contextSpan != null) {
                            definition.contextSpan = h.getTextSpan(document, definition.contextSpan, nodeAtCursor.node);
                        }
                        definition.fileName = document.container.fsPath;
                        definitions.push(definition);
                    }
                }
            }
            else {
                definitions.push(definition);
            }
        });
        result.definitions = definitions;
        if (!isTextSpanSet) {
            if (isSimpleExpressionNode(nodeAtCursor.node)) {
                result.textSpan = h.getTextSpan(document, result.textSpan, nodeAtCursor.node);
            }
            else {
                const textSpan = getTextSpan(document, position);
                if (textSpan != null)
                    result.textSpan = textSpan;
            }
        }
        return result;
    },
};
function getTextSpan(document, position) {
    if (document.ast == null)
        return null;
    const targetNode = findTemplateNodeAt(document.ast, position);
    if ((targetNode === null || targetNode === void 0 ? void 0 : targetNode.node) == null)
        return null;
    if (isAttributeNode(targetNode.node)) {
        return {
            start: targetNode.node.loc.start.offset,
            length: targetNode.node.name.length,
        };
    }
    else if (isElementNode(targetNode.node)) {
        return {
            start: targetNode.node.loc.start.offset + 1,
            length: targetNode.node.tag.length,
        };
    }
    else {
        return {
            start: targetNode.node.loc.start.offset,
            length: targetNode.node.loc.source.length,
        };
    }
}

const GOTO_PROVIDERS = [GotoScriptDefinitions, GotoExternalComponentProp];

const RenameExpression = {
    version: '*',
    name: 'expression',
    canRename(config, fileName, position, preferences) {
        const { node, document } = config.helpers.findTemplateNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node) && document != null) {
            const mappedPosition = document.getGeneratedOffsetAt(position);
            if ((mappedPosition === null || mappedPosition === void 0 ? void 0 : mappedPosition.offset) != null) {
                const result = config.service.getRenameInfo(fileName, mappedPosition.offset, preferences);
                if (result.canRename) {
                    if (result.displayName === '$event') {
                        return {
                            canRename: false,
                            localizedErrorMessage: '$event is builtin variable, it cannot be renamed.',
                        };
                    }
                    result.triggerSpan = config.helpers.getTextSpan(document, result.triggerSpan);
                }
                return result;
            }
            return {
                canRename: false,
                localizedErrorMessage: 'Cannot find mapped position in render function.',
            };
        }
    },
    applyRename(config, fileName, position, findInStrings, findInComments) {
        var _a;
        const { node, document } = config.helpers.findTemplateNodeAtPosition(fileName, position);
        if (isSimpleExpressionNode(node) && document != null) {
            const mappedPosition = document.getGeneratedOffsetAt(position);
            if ((mappedPosition === null || mappedPosition === void 0 ? void 0 : mappedPosition.offset) != null) {
                const result = (_a = config.service
                    .findRenameLocations(fileName, mappedPosition.offset, findInStrings, findInComments)) === null || _a === void 0 ? void 0 : _a.slice();
                return result;
            }
        }
    },
    applyFileRename() {
        return undefined;
    },
};

const RenameElementTag = {
    version: '*',
    name: 'element-tag',
    canRename(config, fileName, position) {
        const { node } = config.helpers.findTemplateNodeAtPosition(fileName, position);
        if (isPlainElementNode(node) && isPositionInTagName$1(position, node)) {
            return {
                canRename: true,
                displayName: node.tag,
                fullDisplayName: node.tag,
                kind: config.context.typescript.ScriptElementKind.unknown,
                kindModifiers: 'elementTagName',
                triggerSpan: {
                    start: node.loc.start.offset + 1,
                    length: node.tag.length,
                },
            };
        }
    },
    applyRename(config, fileName, position) {
        const { node, document } = config.helpers.findTemplateNodeAtPosition(fileName, position);
        if (isPlainElementNode(node) &&
            isPositionInTagName$1(position, node) &&
            document != null) {
            const locations = [
                {
                    fileName: document.container.fsPath,
                    textSpan: {
                        start: node.loc.start.offset + 1,
                        length: node.tag.length,
                    },
                },
            ];
            if (!node.isSelfClosing) {
                locations.push({
                    fileName: document.container.fsPath,
                    textSpan: {
                        start: node.loc.start.offset +
                            node.loc.source.lastIndexOf('</' + node.tag) +
                            2,
                        length: node.tag.length,
                    },
                });
            }
            return locations;
        }
    },
    applyFileRename() {
        return undefined;
    },
};
function isPositionInTagName$1(position, node) {
    return (
    // In start tag.
    position <= node.loc.start.offset + node.tag.length ||
        // In end tag.
        (!node.isSelfClosing &&
            node.loc.start.offset + node.loc.source.lastIndexOf('</') <= position));
}

const RENAME_PROVIDERS = [RenameExpression, RenameComponentTag, RenameElementTag];

function createTemplateLanguageServer(config) {
    const { helpers: h, context, service } = config;
    const choose = (fileName) => {
        var _a, _b, _c;
        try {
            (_a = service.getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(fileName); // This should throw if {fileName} is not part of program.
            return service;
        }
        catch (_d) {
            return ((_c = (_b = context.projectService
                .getDefaultProjectForFile(context.typescript.server.toNormalizedPath(fileName), false)) === null || _b === void 0 ? void 0 : _b.getLanguageService()) !== null && _c !== void 0 ? _c : service);
        }
    };
    return wrapInTrace('TemplateLanguageServer', {
        ...noop,
        getCompletionsAtPosition(fileName, position, options) {
            const result = {
                isGlobalCompletion: false,
                isMemberCompletion: false,
                isNewIdentifierLocation: false,
                entries: [],
            };
            for (const provider of TEMPLATE_COMPLETION_PROVIDERS) {
                const providerResults = provider.getCompletionsAtPosition(config, fileName, position, options);
                if (isNotNull(providerResults)) {
                    result.entries.push(...providerResults.entries);
                    if (isNotNull(providerResults.metadata)) {
                        context.log(`@@DEBUG Found some completion metadata: ${inspect(providerResults.metadata)}`);
                    }
                }
            }
            return result;
        },
        getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
            for (const provider of TEMPLATE_COMPLETION_PROVIDERS) {
                const result = provider.getCompletionEntryDetails(config, fileName, position, entryName, formatOptions, source, preferences);
                if (isNotNull(result))
                    return result;
            }
        },
        getCompletionEntrySymbol(fileName, position, name, source) {
            for (const provider of TEMPLATE_COMPLETION_PROVIDERS) {
                const result = provider.getCompletionEntrySymbol(config, fileName, position, name, source);
                if (isNotNull(result))
                    return result;
            }
        },
        getSignatureHelpItems(fileName, position, options) {
            const document = h.getRenderDoc(fileName);
            const loc = document === null || document === void 0 ? void 0 : document.getGeneratedOffsetAt(position);
            if (loc != null && document != null) {
                return choose(document.fsPath).getSignatureHelpItems(document.fsPath, loc.offset, options);
            }
        },
        getQuickInfoAtPosition(fileName, position) {
            var _a;
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return;
            const loc = document.getGeneratedOffsetAt(position);
            if (loc == null)
                return;
            const nodeAtCursor = h.findTemplateNodeAtPosition(document.fsPath, position);
            const result = choose(document.fsPath).getQuickInfoAtPosition(fileName, loc.offset);
            if (result != null) {
                const textSpan = h.getTextSpan(document, result.textSpan, nodeAtCursor.node);
                if (textSpan != null) {
                    result.textSpan = textSpan;
                }
                else {
                    return;
                }
                (_a = result.displayParts) === null || _a === void 0 ? void 0 : _a.forEach((displayPart) => {
                    if (displayPart.text === 'JSX attribute') {
                        displayPart.text = 'prop'; // TODO: Maybe event or prop or attribute?
                    }
                });
            }
            return result;
        },
        getSemanticDiagnostics(fileName) {
            var _a;
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return [];
            const diagnostics = choose(document.fsPath)
                .getSemanticDiagnostics(document.fsPath)
                .map((diagnostic) => {
                var _a;
                if (diagnostic.file != null &&
                    diagnostic.file.fileName !== document.fsPath) {
                    return diagnostic;
                }
                if (diagnostic.start != null) {
                    if (document.isInGeneratedRange(diagnostic.start)) {
                        const position = document.getOriginalOffsetAt(diagnostic.start);
                        diagnostic.start = (_a = position === null || position === void 0 ? void 0 : position.offset) !== null && _a !== void 0 ? _a : diagnostic.start;
                        return diagnostic;
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return diagnostic;
                }
            })
                .filter(isNotNull);
            // TODO: Cache it.
            const project = context.getVueProjectForFile(document.container.fsPath);
            if (document.ast != null && (project === null || project === void 0 ? void 0 : project.kind) === 'inferred') {
                const info = h.getComponentInfo(document.container);
                const localComponentNames = new Set(info.components.flatMap((component) => component.aliases));
                const globalComponentNames = new Set(project.globalComponents.flatMap((component) => component.aliases));
                const projectComponentNames = new Set(project.components.flatMap((component) => component.aliases));
                const projectComponents = project.components;
                const file = (_a = choose(document.fsPath)
                    .getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(document.fsPath);
                traverseFast(document.ast, (node) => {
                    if (isComponentNode(node)) {
                        if (!localComponentNames.has(node.tag) &&
                            !globalComponentNames.has(node.tag) &&
                            projectComponentNames.has(node.tag)) {
                            const components = projectComponents.filter((component) => component.aliases.includes(node.tag));
                            let messageText = `The component '${node.tag}' is inferred as global component. It may not be available at runtime.`;
                            const relatedInformation = [];
                            if (!components[0].source.moduleName.endsWith('.vue')) {
                                messageText = `The component '${node.tag}' is found in '${components[0].source.moduleName}' and it inferred as global component. It may not be available at runtime.`;
                            }
                            if (components.length > 1) {
                                components.forEach((component) => {
                                    relatedInformation.push({
                                        code: 59002,
                                        category: context.typescript.DiagnosticCategory.Warning,
                                        messageText: `Found in '${component.source.moduleName}'`,
                                        file: undefined,
                                        start: undefined,
                                        length: undefined,
                                    });
                                });
                            }
                            diagnostics.push({
                                category: context.typescript.DiagnosticCategory.Warning,
                                code: 59001,
                                start: node.loc.start.offset + 1,
                                length: node.tag.length,
                                messageText,
                                file,
                                relatedInformation,
                            });
                        }
                    }
                });
            }
            if (document.container.descriptor.template != null) {
                const start = document.container.descriptor.template.loc.start.offset;
                const end = document.container.descriptor.template.loc.end.offset;
                return diagnostics.filter((item) => item.start == null ||
                    item.length == null ||
                    (start <= item.start && item.start + item.length <= end));
            }
            return diagnostics;
        },
        getSuggestionDiagnostics(fileName) {
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return [];
            const diagnostics = choose(document.fsPath)
                .getSuggestionDiagnostics(document.fsPath)
                .map((diagnostic) => {
                var _a;
                if (diagnostic.start != null) {
                    if (document.isInGeneratedRange(diagnostic.start)) {
                        const position = document.getOriginalOffsetAt(diagnostic.start);
                        diagnostic.start = (_a = position === null || position === void 0 ? void 0 : position.offset) !== null && _a !== void 0 ? _a : diagnostic.start;
                        return diagnostic;
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return diagnostic;
                }
            })
                .filter(isNotNull);
            if (document.container.descriptor.template != null) {
                const start = document.container.descriptor.template.loc.start.offset;
                const end = document.container.descriptor.template.loc.end.offset;
                return diagnostics.filter((item) => item.start == null ||
                    item.length == null ||
                    (start <= item.start && item.start + item.length <= end));
            }
            return diagnostics;
        },
        getSyntacticDiagnostics(fileName) {
            var _a;
            const document = h.getRenderDoc(fileName);
            if (document == null)
                return [];
            const diagnostics = choose(document.fsPath)
                .getSyntacticDiagnostics(document.fsPath)
                .map((diagnostic) => {
                var _a;
                if (diagnostic.start != null) {
                    if (document.isInGeneratedRange(diagnostic.start)) {
                        const position = document.getOriginalOffsetAt(diagnostic.start);
                        diagnostic.start = (_a = position === null || position === void 0 ? void 0 : position.offset) !== null && _a !== void 0 ? _a : diagnostic.start;
                        return diagnostic;
                    }
                    else {
                        return null;
                    }
                }
                else {
                    return diagnostic;
                }
            })
                .filter(isNotNull);
            if (document.parserErrors.length > 0) {
                const sourceFile = (_a = choose(document.fsPath)
                    .getProgram()) === null || _a === void 0 ? void 0 : _a.getSourceFile(document.fsPath);
                const block = document.container.getBlock({ type: 'template' });
                if (sourceFile != null && block != null) {
                    const start = block.loc.start.offset;
                    const length = block.loc.end.offset - start;
                    document.parserErrors.forEach((error) => {
                        var _a;
                        diagnostics.push({
                            category: context.typescript.DiagnosticCategory.Error,
                            code: 50000 + error.code,
                            file: sourceFile,
                            source: (_a = error.loc) === null || _a === void 0 ? void 0 : _a.source,
                            start: error.loc != null ? error.loc.start.offset : start,
                            length: error.loc != null
                                ? error.loc.end.offset - error.loc.start.offset
                                : length,
                            messageText: error.message,
                        });
                    });
                }
            }
            if (document.container.descriptor.template != null) {
                const start = document.container.descriptor.template.loc.start.offset;
                const end = document.container.descriptor.template.loc.end.offset;
                return diagnostics.filter((item) => item.start == null ||
                    item.length == null ||
                    (start <= item.start && item.start + item.length <= end));
            }
            return diagnostics;
        },
        getRenameInfo(fileName, position, preferences = {}) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.canRename(config, fileName, position, preferences);
                if (result != null) {
                    context.log(`@@DEBUG found getRenameInfo using "${provider.name}"`);
                    return result;
                }
            }
            return {
                canRename: false,
                localizedErrorMessage: 'You cannot rename this element.',
            };
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.applyRename(config, fileName, position, findInStrings, findInComments);
                if (result != null) {
                    context.log(`@@DEBUG found findRenameLocations using "${provider.name}"`);
                    return result;
                }
            }
        },
        getEditsForFileRenameIn(fileName, oldFilePath, newFilePath) {
            const fileTextChanges = [];
            for (const provider of RENAME_PROVIDERS) {
                const result = provider.applyFileRename(config, fileName, oldFilePath, newFilePath, {}, {});
                if (result != null)
                    fileTextChanges.push(...result);
            }
            return fileTextChanges;
        },
        getApplicableRefactors(fileName, position, preferences = {}) {
            const refactors = [];
            for (const provider of REFACTOR_PROVIDERS) {
                const result = provider.findRefactors(config, fileName, position, preferences);
                if (result != null)
                    refactors.push(...result);
            }
            return refactors;
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences = {}) {
            const provider = REFACTOR_PROVIDERS.find((provider) => provider.name === refactorName);
            if (provider != null) {
                context.log(`@@DEBUG: Using ${provider.name}`);
                return provider.applyRefactor(config, fileName, formatOptions, positionOrRange, refactorName, actionName, preferences);
            }
        },
        getDefinitionAndBoundSpan(fileName, position) {
            for (const provider of GOTO_PROVIDERS) {
                const result = provider.getDefinitionAndBoundSpan(config, fileName, position);
                if (result != null)
                    return result;
            }
        },
        getJsxClosingTagAtPosition(fileName, position) {
            const { node } = h.findTemplateNodeAtPosition(fileName, position);
            if (isElementNode(node)) {
                return { newText: `</${node.tag}>` };
            }
        },
    });
}

function createVueLanguageServer(options) {
    const template = createTemplateLanguageServer(options);
    const { helpers: h, service, context } = options;
    const script = wrapInTrace('ScriptLanguageServer', service);
    function choose(document) {
        if (h.isRenderFunctionDocument(document)) {
            return template;
        }
        return script;
    }
    return wrapInTrace('VueLanguageServer', {
        ...noop,
        getSemanticDiagnostics(fileName) {
            if (!h.isFeatureEnabled('diagnostics', 'semantic')) {
                return [];
            }
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document != null) {
                const selectors = ['script', '_render', 'scriptSetup'];
                selectors.forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual != null) {
                        const results = choose(virtual).getSemanticDiagnostics(virtual.fsPath);
                        diagnostics.push(...results);
                    }
                });
            }
            return diagnostics;
        },
        getSuggestionDiagnostics(fileName) {
            if (!h.isFeatureEnabled('diagnostics', 'suggestion')) {
                return [];
            }
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document != null) {
                const selectors = ['script', '_render', 'scriptSetup'];
                selectors.forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual != null)
                        diagnostics.push(...choose(virtual).getSuggestionDiagnostics(virtual.fsPath));
                });
            }
            return diagnostics;
        },
        getSyntacticDiagnostics(fileName) {
            if (!h.isFeatureEnabled('diagnostics', 'syntactic')) {
                return [];
            }
            const document = h.getVueDocument(fileName);
            const diagnostics = [];
            if (document != null) {
                const selectors = ['script', '_render', 'scriptSetup'];
                selectors.forEach((selector) => {
                    const virtual = document.getDocument(selector);
                    if (virtual != null)
                        diagnostics.push(...choose(virtual).getSyntacticDiagnostics(virtual.fsPath));
                });
            }
            return diagnostics;
        },
        organizeImports(scope, formatOptions, preferences) {
            var _a;
            if (!h.isFeatureEnabled('organizeImports'))
                return [];
            const document = h.getVueDocument(scope.fileName);
            if (document != null) {
                const virtual = (_a = document.getDocument('script')) !== null && _a !== void 0 ? _a : document.getDocument('scriptSetup');
                if (virtual != null) {
                    return script.organizeImports({ ...scope, fileName: virtual.fsPath }, formatOptions, preferences);
                }
            }
            return [];
        },
        getQuickInfoAtPosition(fileName, position) {
            if (!h.isFeatureEnabled('quickInfo'))
                return;
            // TODO: Provide better quick info for components and props.
            const document = h.getDocumentAt(fileName, position);
            if (document != null)
                return choose(document).getQuickInfoAtPosition(document.fsPath, position);
        },
        getRenameInfo(fileName, position, options) {
            if (!h.isFeatureEnabled('rename')) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Rename feature disabled.',
                };
            }
            const document = h.getDocumentAt(fileName, position);
            if (document == null) {
                return {
                    canRename: false,
                    localizedErrorMessage: 'Cannot rename this entity.',
                };
            }
            const result = choose(document).getRenameInfo(document.fsPath, position, options);
            return result;
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            var _a, _b, _c, _d;
            if (!h.isFeatureEnabled('rename'))
                return;
            const document = h.getDocumentAt(fileName, position);
            if (document == null)
                return;
            const locations = (_a = choose(document)
                .findRenameLocations(document.fsPath, position, findInStrings, findInComments)) === null || _a === void 0 ? void 0 : _a.slice();
            if (locations != null && locations.length > 0) {
                const result = choose(document).getRenameInfo(document.fsPath, position, {
                    allowRenameOfImportPath: false,
                });
                if (result.canRename) {
                    const info = h.getComponentInfo(document.container);
                    if (!h.isRenderFunctionDocument(document)) {
                        // Maybe component got renamed
                        const component = info.components.find((component) => component.name === result.displayName);
                        if (component != null &&
                            locations.some((loc) => isSpanInSourceRange(loc.textSpan, component.loc) ||
                                isSpanInSourceRange(loc.textSpan, component.source.loc))) {
                            const doc = h.getRenderDoc(document.fsPath);
                            if ((doc === null || doc === void 0 ? void 0 : doc.ast) != null) {
                                getChangesForComponentTagRename(doc.ast, component).forEach((textSpan) => {
                                    locations.push({ fileName, textSpan });
                                });
                            }
                        }
                        else if ((info.options != null &&
                            isSpanInSourceRange(result.triggerSpan, info.options.loc)) ||
                            (((_b = info.fnSetupOption) === null || _b === void 0 ? void 0 : _b.return) != null &&
                                isSpanInSourceRange(result.triggerSpan, info.fnSetupOption.return.loc))) {
                            if (isSimpleIdentifier(result.displayName)) {
                                const property = Object.entries((_d = (_c = info.options) === null || _c === void 0 ? void 0 : _c.properties) !== null && _d !== void 0 ? _d : {
                                    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                                    setup: info.fnSetupOption.return,
                                }).find(([_, value]) => isSpanInSourceRange(result.triggerSpan, value.loc));
                                if (property != null) {
                                    switch (property[0]) {
                                        case 'data':
                                        case 'methods':
                                        case 'computed':
                                        case 'props':
                                        case 'emits':
                                        case 'setup':
                                            {
                                                const render = h.getRenderDoc(fileName);
                                                if ((render === null || render === void 0 ? void 0 : render.ast) != null) {
                                                    if (render.ast.scope.identifiers.includes(result.displayName)) {
                                                        // Find an expression using this identifiers.
                                                        let isDone = false;
                                                        traverseEvery(render.ast, (node) => {
                                                            if (isDone)
                                                                return false;
                                                            if (isSimpleExpressionNode(node) &&
                                                                node.scope.identifiers.includes(result.displayName)) {
                                                                const binding = node.scope.getBinding(result.displayName);
                                                                if (binding == null) {
                                                                    isDone = true;
                                                                    context.log('@@DEBUG found expression usage in template');
                                                                    const additionalLocations = template.findRenameLocations(document.container.getDocumentFileName(RENDER_SELECTOR), node.loc.start.offset +
                                                                        node.content.indexOf(result.displayName) +
                                                                        1, findInStrings, findInComments);
                                                                    if (additionalLocations != null) {
                                                                        context.log('@@DEBUG found additional change locations');
                                                                        locations.push(...additionalLocations);
                                                                    }
                                                                    return false; // stop
                                                                }
                                                            }
                                                            return true;
                                                        });
                                                    }
                                                }
                                            }
                                            break;
                                    }
                                }
                            }
                        }
                    }
                    else {
                        if (isSimpleIdentifier(result.displayName)) {
                            const source = info.identifierSource[result.displayName.trim()];
                            if (source != null) {
                                const scriptFileName = document.container.descriptor.scriptSetup != null
                                    ? document.container.getDocumentFileName(SCRIPT_SETUP_BLOCK_SELECTOR)
                                    : document.container.getDocumentFileName(SCRIPT_BLOCK_SELECTOR);
                                const additionalLocations = script.findRenameLocations(scriptFileName, source.loc.start.offset, findInStrings, findInComments);
                                if (additionalLocations != null) {
                                    locations.push(...additionalLocations);
                                }
                            }
                        }
                    }
                }
            }
            return locations;
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            if (!h.isFeatureEnabled('rename'))
                return [];
            const document = h.getVueDocument(oldFilePath);
            const fileTextChanges = [];
            const visited = new Set();
            if (document != null) {
                const component = document.getDocument(MODULE_SELECTOR);
                const currentChanges = script.getEditsForFileRename(component.fsPath, isVueFile(newFilePath)
                    ? newFilePath + VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR
                    : newFilePath, formatOptions, preferences);
                fileTextChanges.push(...currentChanges);
                currentChanges.forEach((item) => {
                    var _a;
                    if (isVirtualFile(item.fileName) || isVueFile(item.fileName)) {
                        const render = (_a = h
                            .getVueDocument(item.fileName)) === null || _a === void 0 ? void 0 : _a.getDocument('_render');
                        if (render != null && !visited.has(render.fsPath)) {
                            visited.add(render.fsPath);
                            fileTextChanges.push(...template.getEditsForFileRenameIn(render.fsPath, oldFilePath, newFilePath));
                        }
                    }
                });
            }
            return fileTextChanges;
        },
        getApplicableRefactors(fileName, position, preferences) {
            if (!h.isFeatureEnabled('refactor'))
                return [];
            const document = h.getDocumentAt(fileName, position);
            const result = document != null
                ? choose(document).getApplicableRefactors(document.fsPath, position, preferences)
                : [];
            if (document != null && !h.isRenderFunctionDocument(document)) {
                SCRIPT_REFACTOR_PROVIDERS.forEach((provider) => {
                    result.push(...provider.findRefactors(options, document.fsPath, position, preferences));
                });
            }
            return result;
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
            if (!h.isFeatureEnabled('refactor'))
                return;
            const document = h.getDocumentAt(fileName, positionOrRange);
            if (document != null) {
                const provider = SCRIPT_REFACTOR_PROVIDERS.find((provider) => provider.name === refactorName);
                if (provider != null) {
                    return provider.applyRefactor(options, document.fsPath, formatOptions, positionOrRange, refactorName, actionName, preferences);
                }
                const result = choose(document).getEditsForRefactor(document.fsPath, formatOptions, positionOrRange, refactorName, actionName, preferences);
                if ((result === null || result === void 0 ? void 0 : result.renameLocation) != null) {
                    const isExtractComponent = refactorName === REFACTORS.EXTRACT_COMPONENT;
                    const renameLocation = result.renameLocation;
                    let delta = 0;
                    result.edits.forEach((edit) => {
                        if (edit.fileName === fileName) {
                            edit.textChanges.forEach((textChange) => {
                                if (textChange.span.start + textChange.span.length <
                                    renameLocation &&
                                    (!isExtractComponent ||
                                        // For extract-component, renameLocation should be generated import statement.
                                        textChange.span.start -
                                            textChange.span.length +
                                            textChange.newText.length <
                                            renameLocation)) {
                                    delta -= textChange.span.length;
                                    delta += textChange.newText.length;
                                }
                            });
                        }
                    });
                    result.renameLocation = renameLocation + delta;
                }
                return result;
            }
        },
        getDocumentHighlights(fileName, position, filesToSearch) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null && !h.isRenderFunctionDocument(document)) {
                return script.getDocumentHighlights(document.fsPath, position, filesToSearch
                    .map((fileName) => {
                    var _a;
                    if (isVueFile(fileName)) {
                        const document = h.getVueDocument(fileName);
                        if (document == null)
                            return;
                        const script = (_a = document.getDocument('script')) !== null && _a !== void 0 ? _a : document.getDocument('scriptSetup');
                        if (script == null)
                            return;
                        return script.fsPath;
                    }
                    return fileName;
                })
                    .filter(Boolean));
            }
        },
        getCompletionsAtPosition(fileName, position, options) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getCompletionsAtPosition(document.fsPath, position, options);
            }
        },
        getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getCompletionEntryDetails(document.fsPath, position, entryName, formatOptions, source, preferences);
            }
        },
        getSignatureHelpItems(fileName, position, options) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getSignatureHelpItems(document.fsPath, position, options);
            }
        },
        getCompletionEntrySymbol(fileName, position, name, source) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getCompletionEntrySymbol(document.fsPath, position, name, source);
            }
        },
        getDefinitionAtPosition(fileName, position) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getDefinitionAtPosition(document.fsPath, position);
            }
        },
        getDefinitionAndBoundSpan(fileName, position) {
            if (!h.isFeatureEnabled('goto'))
                return;
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getDefinitionAndBoundSpan(document.fsPath, position);
            }
        },
        getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
            const document = h.getDocumentAt(fileName, start);
            if (document != null) {
                return choose(document).getCodeFixesAtPosition(document.fsPath, start, end, errorCodes, formatOptions, preferences);
            }
            return [];
        },
        getJsxClosingTagAtPosition(fileName, position) {
            const document = h.getDocumentAt(fileName, position);
            if (document != null) {
                return choose(document).getJsxClosingTagAtPosition(document.fsPath, position);
            }
        },
    });
}

const VUE_LANGUAGE_SERVER = Symbol('Vue Language Server');
class RoutingLanguageServer {
    constructor(context) {
        this.context = context;
    }
    decorate(languageService) {
        if (VUE_LANGUAGE_SERVER in languageService) {
            return languageService;
        }
        const proxy = createLanguageServiceRouter({
            context: this.context,
            service: languageService,
            helpers: createServerHelper(this.context, languageService),
        });
        // @ts-expect-error
        proxy[VUE_LANGUAGE_SERVER] = true;
        return proxy;
    }
}
function createLanguageServiceRouter(config) {
    const vue = createVueLanguageServer(config);
    const virtual = createVirtualLanguageServer(config);
    const ts = config.service;
    function choose(fileName) {
        if (fileName == null)
            return ts;
        return fileName.startsWith('^vue:')
            ? virtual
            : isVueFile(fileName) || isVirtualFile(fileName)
                ? vue
                : ts;
    }
    function getTextSpan(document, span) {
        if (config.helpers.isRenderFunctionDocument(document)) {
            const result = document.getOriginalOffsetAt(span.start);
            if (result != null) {
                return {
                    start: result.offset,
                    length: Math.min(span.length, result.length),
                };
            }
            return null;
        }
        return span;
    }
    const VIRTUAL_FILE_SUFFIX_RE = new RegExp(`(?<=\\.vue)${VIRTUAL_FILENAME_SEPARATOR}([A-Za-z_][A-Za-z0-9_-]*)(\\.[jt]sx?)?`, 'g');
    function getComponentName$1(fileName) {
        if (fileName == null || !isVueFile(fileName))
            return;
        return getComponentName(fileName);
    }
    const COMPONENT_TYPE_RE = /'ComponentPublicInstance<.*?ComponentOptionsBase<...>>'/g;
    const REPLACE = {
        virtualFile: (messageText) => messageText.replace(VIRTUAL_FILE_SUFFIX_RE, ''),
        componentType: (fileName, messageText) => {
            const component = getComponentName$1(fileName);
            return component != null
                ? messageText.replace(COMPONENT_TYPE_RE, `vue component '${component}'`)
                : messageText;
        },
    };
    function applyReplacements(fileName, messageText) {
        if (typeof messageText === 'string') {
            return REPLACE.componentType(fileName, REPLACE.virtualFile(messageText));
        }
        else {
            messageText.messageText = applyReplacements(fileName, messageText.messageText);
            return messageText;
        }
    }
    function isVirtualSourceFile(file) {
        return file != null && isVirtualFile(file.fileName);
    }
    const proxy = wrapInTrace('RoutingLanguageServer', {
        ...config.service,
        getSyntacticDiagnostics(fileName) {
            const diagnostics = choose(fileName).getSyntacticDiagnostics(fileName);
            return diagnostics
                .map((diagnostic) => {
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation != null) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (info.file != null && isVirtualFile(info.file.fileName)) {
                            info.file = {
                                ...info.file,
                                fileName: getContainingFile(info.file.fileName),
                            };
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getSemanticDiagnostics(fileName) {
            config.context.disposeUnusedProjects();
            const diagnostics = choose(fileName).getSemanticDiagnostics(fileName);
            return diagnostics
                .map((diagnostic) => {
                if (isVirtualSourceFile(diagnostic.file)) {
                    diagnostic.file = {
                        ...diagnostic.file,
                        fileName: getContainingFile(diagnostic.file.fileName),
                    };
                }
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation != null) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (info.file != null && isVirtualFile(info.file.fileName)) {
                            info.file = {
                                ...info.file,
                                fileName: getContainingFile(info.file.fileName),
                            };
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getSuggestionDiagnostics(fileName) {
            const diagnostics = choose(fileName).getSuggestionDiagnostics(fileName);
            return diagnostics
                .map((diagnostic) => {
                diagnostic.messageText = applyReplacements(fileName, diagnostic.messageText);
                if (diagnostic.relatedInformation != null) {
                    diagnostic.relatedInformation = diagnostic.relatedInformation.map((info) => {
                        info.messageText = applyReplacements(fileName, info.messageText);
                        if (info.file != null && isVirtualFile(info.file.fileName)) {
                            info.file = {
                                ...info.file,
                                fileName: getContainingFile(info.file.fileName),
                            };
                        }
                        return info;
                    });
                }
                return diagnostic;
            })
                .filter(isNotNull);
        },
        getCompilerOptionsDiagnostics() {
            return ts.getCompilerOptionsDiagnostics();
        },
        getEncodedSemanticClassifications(fileName, span) {
            return ts.getEncodedSemanticClassifications(fileName, span);
        },
        getEncodedSyntacticClassifications(fileName, span) {
            return ts.getEncodedSyntacticClassifications(fileName, span);
        },
        getCompletionsAtPosition(fileName, position, options) {
            const completions = choose(fileName).getCompletionsAtPosition(fileName, position, options);
            if ((completions === null || completions === void 0 ? void 0 : completions.entries) != null) {
                const additionalEntries = [];
                completions.entries = completions.entries
                    .filter((entry) => {
                    if (entry.source != null && isVirtualFile(entry.source)) {
                        return isVirtualFileOfType(entry.source, '_module');
                    }
                    return true;
                })
                    .map((entry) => {
                    if (entry.source != null && isVirtualFile(entry.source)) {
                        entry.source = getContainingFile(entry.source);
                        const componentName = getComponentName$1(entry.source);
                        if (componentName === entry.name) {
                            additionalEntries.push({
                                ...entry,
                                name: kebabCase(entry.name),
                                isRecommended: undefined,
                                insertText: entry.insertText != null
                                    ? kebabCase(entry.insertText)
                                    : undefined,
                            });
                        }
                    }
                    return entry;
                });
                completions.entries.push(...additionalEntries);
            }
            return completions;
        },
        getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences) {
            var _a, _b, _c;
            let details;
            if (source != null && isVueFile(source)) {
                // -> Importing from a .vue file
                let newSource = source;
                let newEntryName = entryName;
                const componentName = getComponentName(source);
                if (newEntryName.includes('-') &&
                    getComponentName(newEntryName) === componentName) {
                    newEntryName = componentName;
                }
                // -> Rewrite source to _module virtual file.
                newSource = source + VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR;
                details = choose(fileName).getCompletionEntryDetails(fileName, position, newEntryName, formatOptions, newSource, preferences);
                const isComponentImport = newEntryName === componentName;
                if (isComponentImport) {
                    const documentAtCursor = config.helpers.getDocumentAt(fileName, position);
                    (_a = details === null || details === void 0 ? void 0 : details.codeActions) === null || _a === void 0 ? void 0 : _a.forEach((codeAction) => {
                        codeAction.description = codeAction.description.replace(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR, '');
                        codeAction.changes.forEach((change) => {
                            var _a;
                            const document = config.helpers.getVueDocument(change.fileName);
                            if (change.textChanges.length > 0 &&
                                document != null &&
                                config.helpers.isRenderFunctionDocument(documentAtCursor)) {
                                const block = (_a = document.descriptor.scriptSetup) !== null && _a !== void 0 ? _a : document.descriptor.script;
                                if (block == null) {
                                    const project = config.context.getVueProjectForFile(fileName, true);
                                    change.textChanges = registerLocalComponentWithSource(document, config.helpers.getComponentInfo(document), { moduleName: source, localName: newEntryName }, // <- create fake "source" as "importStatement" is available
                                    project.config.preferences.script, change.textChanges[0].newText);
                                }
                                else {
                                    change.textChanges.forEach((textChange) => {
                                        // We need to rewrite import statement to start of script block of .vue file
                                        if (block != null) {
                                            // Add import statement to start of <script> or <script setup> block
                                            textChange.span.start = block.loc.start.offset + 1;
                                        }
                                    });
                                    change.textChanges = [
                                        ...change.textChanges,
                                        ...registerComponentAPI(document, config.helpers.getComponentInfo(document), 'components', newEntryName, newEntryName).changes,
                                    ];
                                }
                            }
                        });
                    });
                }
            }
            else {
                details = choose(fileName).getCompletionEntryDetails(fileName, position, entryName, formatOptions, source, preferences);
            }
            (_b = details === null || details === void 0 ? void 0 : details.codeActions) === null || _b === void 0 ? void 0 : _b.forEach((codeAction) => {
                codeAction.changes.forEach((change) => {
                    if (isVirtualFile(change.fileName)) {
                        change.fileName = getContainingFile(change.fileName);
                    }
                    change.textChanges.forEach((textChange) => {
                        textChange.newText = textChange.newText.replace(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR, '');
                    });
                });
            });
            (_c = details === null || details === void 0 ? void 0 : details.source) === null || _c === void 0 ? void 0 : _c.forEach((item) => {
                if (isVirtualFile(item.text)) {
                    item.text = getContainingFile(item.text);
                }
            });
            return details;
        },
        getCompletionEntrySymbol(fileName, position, name, source) {
            return choose(fileName).getCompletionEntrySymbol(fileName, position, name, source);
        },
        getQuickInfoAtPosition(fileName, position) {
            const info = choose(fileName).getQuickInfoAtPosition(fileName, position);
            if ((info === null || info === void 0 ? void 0 : info.displayParts) != null) {
                info.displayParts = info.displayParts
                    .map((part) => {
                    part.text = applyReplacements(fileName, part.text);
                    return part;
                })
                    .filter(isNotNull);
            }
            if ((info === null || info === void 0 ? void 0 : info.documentation) != null) {
                info.documentation = info.documentation
                    .map((part) => {
                    part.text = applyReplacements(fileName, part.text);
                    return part;
                })
                    .filter(isNotNull);
            }
            return info;
        },
        getNameOrDottedNameSpan(fileName, startPos, endPos) {
            return choose(fileName).getNameOrDottedNameSpan(fileName, startPos, endPos);
        },
        getBreakpointStatementAtPosition(fileName, position) {
            return choose(fileName).getBreakpointStatementAtPosition(fileName, position);
        },
        getSignatureHelpItems(fileName, position, options) {
            return choose(fileName).getSignatureHelpItems(fileName, position, options);
        },
        getRenameInfo(fileName, position, options) {
            const result = choose(fileName).getRenameInfo(fileName, position, options);
            if (result.canRename) {
                if (result.fileToRename != null &&
                    isVirtualFile(result.fileToRename)) {
                    result.fileToRename = getContainingFile(result.fileToRename);
                }
                result.displayName = REPLACE.virtualFile(result.displayName);
                result.fullDisplayName = REPLACE.virtualFile(result.fullDisplayName);
                if (result.kind ===
                    config.context.typescript.ScriptElementKind.moduleElement &&
                    result.displayName.endsWith('.vue')) {
                    result.triggerSpan.length -= 4;
                }
            }
            return result;
        },
        findRenameLocations(fileName, position, findInStrings, findInComments) {
            var _a;
            const result = (_a = choose(fileName)
                .findRenameLocations(fileName, position, findInStrings, findInComments)) === null || _a === void 0 ? void 0 : _a.map((item) => {
                if (isVirtualFile(item.fileName)) {
                    const virtual = config.helpers.getDocument(item.fileName);
                    item.fileName = virtual.container.fsPath;
                    const textSpan = getTextSpan(virtual, item.textSpan);
                    if (textSpan == null)
                        return;
                    item.textSpan = textSpan;
                    if (item.contextSpan != null) {
                        const contextSpan = getTextSpan(virtual, item.contextSpan);
                        if (contextSpan == null)
                            return;
                        item.contextSpan = contextSpan;
                    }
                }
                return item;
            }).filter(isNotNull);
            return result;
        },
        getSmartSelectionRange(fileName, position) {
            return choose(fileName).getSmartSelectionRange(fileName, position);
        },
        getDefinitionAtPosition(fileName, position) {
            const result = choose(fileName).getDefinitionAtPosition(fileName, position);
            if (result != null) {
                result.forEach((definitionInfo) => {
                    if (isVirtualFile(definitionInfo.fileName)) {
                        definitionInfo.fileName = getContainingFile(definitionInfo.fileName);
                        definitionInfo.name = applyReplacements(definitionInfo.fileName, definitionInfo.name);
                    }
                });
            }
            return result;
        },
        getDefinitionAndBoundSpan(fileName, position) {
            const result = choose(fileName).getDefinitionAndBoundSpan(fileName, position);
            if ((result === null || result === void 0 ? void 0 : result.definitions) != null) {
                result.definitions.forEach((definition) => {
                    var _a, _b;
                    if (isVirtualFile(definition.fileName)) {
                        if (config.helpers.isVueModuleFileName(definition.fileName)) {
                            const document = config.helpers.getVueDocument(definition.fileName);
                            if ((document === null || document === void 0 ? void 0 : document.descriptor) != null) {
                                const script = (_a = document.descriptor.scriptSetup) !== null && _a !== void 0 ? _a : document.descriptor.script;
                                if (script != null) {
                                    // TODO: resolve to export default in <script> or <script setup>
                                    definition.name = getComponentName(document.fsPath);
                                    const info = config.helpers.getComponentInfo(document);
                                    definition.textSpan = definition.contextSpan = {
                                        start: script.loc.start.offset,
                                        length: script.loc.end.offset - script.loc.start.offset,
                                    };
                                    if (((_b = info.options) === null || _b === void 0 ? void 0 : _b.loc) != null) {
                                        definition.textSpan = {
                                            start: info.options.loc.start.offset,
                                            length: info.options.loc.end.offset -
                                                info.options.loc.start.offset +
                                                1,
                                        };
                                    }
                                }
                            }
                        }
                        definition.fileName = getContainingFile(definition.fileName);
                    }
                });
            }
            return result;
        },
        getTypeDefinitionAtPosition(fileName, position) {
            return choose(fileName).getTypeDefinitionAtPosition(fileName, position);
        },
        getImplementationAtPosition(fileName, position) {
            return choose(fileName).getImplementationAtPosition(fileName, position);
        },
        getReferencesAtPosition(fileName, position) {
            return choose(fileName).getReferencesAtPosition(fileName, position);
        },
        findReferences(fileName, position) {
            return choose(fileName).findReferences(fileName, position);
        },
        getDocumentHighlights(fileName, position, filesToSearch) {
            const result = choose(fileName).getDocumentHighlights(fileName, position, filesToSearch);
            if (result != null) {
                result.forEach((highlights) => {
                    if (isVirtualFile(fileName)) {
                        highlights.fileName = getContainingFile(highlights.fileName);
                        highlights.highlightSpans.forEach((span) => {
                            if (span.fileName != null && isVirtualFile(span.fileName)) {
                                span.fileName = getContainingFile(span.fileName);
                            }
                        });
                    }
                });
            }
            return result !== null && result !== void 0 ? result : [];
        },
        getNavigateToItems(searchValue, maxResultCount, fileName, excludeDtsFiles) {
            return choose(fileName).getNavigateToItems(searchValue, maxResultCount, fileName, excludeDtsFiles);
        },
        getNavigationBarItems(fileName) {
            return choose(fileName).getNavigationBarItems(fileName);
        },
        getNavigationTree(fileName) {
            return choose(fileName).getNavigationTree(fileName);
        },
        prepareCallHierarchy(fileName, position) {
            return choose(fileName).prepareCallHierarchy(fileName, position);
        },
        provideCallHierarchyIncomingCalls(fileName, position) {
            return choose(fileName).provideCallHierarchyIncomingCalls(fileName, position);
        },
        provideCallHierarchyOutgoingCalls(fileName, position) {
            return choose(fileName).provideCallHierarchyOutgoingCalls(fileName, position);
        },
        getOutliningSpans(fileName) {
            return choose(fileName).getOutliningSpans(fileName);
        },
        getTodoComments(fileName, descriptors) {
            return choose(fileName).getTodoComments(fileName, descriptors);
        },
        getBraceMatchingAtPosition(fileName, position) {
            return choose(fileName).getBraceMatchingAtPosition(fileName, position);
        },
        getIndentationAtPosition(fileName, position, options) {
            return choose(fileName).getIndentationAtPosition(fileName, position, options);
        },
        getFormattingEditsForRange(fileName, start, end, options) {
            return choose(fileName).getFormattingEditsForRange(fileName, start, end, options);
        },
        getFormattingEditsForDocument(fileName, options) {
            return choose(fileName).getFormattingEditsForDocument(fileName, options);
        },
        getFormattingEditsAfterKeystroke(fileName, position, key, options) {
            return choose(fileName).getFormattingEditsAfterKeystroke(fileName, position, key, options);
        },
        getDocCommentTemplateAtPosition(fileName, position) {
            return choose(fileName).getDocCommentTemplateAtPosition(fileName, position);
        },
        isValidBraceCompletionAtPosition(fileName, position, openingBrace) {
            return choose(fileName).isValidBraceCompletionAtPosition(fileName, position, openingBrace);
        },
        getJsxClosingTagAtPosition(fileName, position) {
            return choose(fileName).getJsxClosingTagAtPosition(fileName, position);
        },
        getSpanOfEnclosingComment(fileName, position, onlyMultiLine) {
            return choose(fileName).getSpanOfEnclosingComment(fileName, position, onlyMultiLine);
        },
        toLineColumnOffset(fileName, position) {
            return config.service.toLineColumnOffset != null
                ? config.service.toLineColumnOffset(fileName, position)
                : { line: 0, character: 0 };
        },
        getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences) {
            return choose(fileName)
                .getCodeFixesAtPosition(fileName, start, end, errorCodes, formatOptions, preferences)
                .filter((fix) => {
                if (fix.fixName === 'import' &&
                    fix.description.includes(VIRTUAL_FILENAME_SEPARATOR)) {
                    return fix.description.includes(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR);
                }
                return true;
            })
                .map((fix) => {
                if (fix.fixName === 'import') {
                    fix.description = fix.description.replace(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR, '');
                    fix.changes.forEach((change) => {
                        change.textChanges.forEach((textChange) => {
                            textChange.newText = textChange.newText.replace(VIRTUAL_FILENAME_SEPARATOR + MODULE_SELECTOR, '');
                        });
                    });
                }
                fix.changes.forEach((change) => {
                    if (isVirtualFile(change.fileName)) {
                        const document = config.helpers.getVueDocument(change.fileName);
                        const fileNameInfo = parseVirtualFileName(change.fileName);
                        change.fileName = getContainingFile(change.fileName);
                        if (document != null && fileNameInfo != null) {
                            const block = document.getBlock(fileNameInfo.selector);
                            if (block != null) {
                                change.textChanges.forEach((textChange) => {
                                    if (textChange.span.start < block.loc.start.offset) {
                                        // Put at beginning of the block.
                                        textChange.span.start = block.loc.start.offset + 1;
                                    }
                                });
                            }
                        }
                    }
                });
                return fix;
            });
        },
        getCombinedCodeFix(scope, fixId, formatOptions, preferences) {
            return config.service.getCombinedCodeFix(scope, fixId, formatOptions, preferences);
        },
        getApplicableRefactors(fileName, positionOrRange, preferences) {
            return choose(fileName).getApplicableRefactors(fileName, positionOrRange, preferences);
        },
        getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences) {
            const result = choose(fileName).getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, preferences);
            if (result != null) {
                result.edits = result.edits.filter((edit) => {
                    var _a;
                    if (isVirtualFile(edit.fileName)) {
                        const info = parseVirtualFileName(edit.fileName);
                        edit.fileName = getContainingFile(edit.fileName);
                        return ['script', 'scriptSetup'].includes((_a = info === null || info === void 0 ? void 0 : info.selector.type) !== null && _a !== void 0 ? _a : '');
                    }
                    else {
                        return true;
                    }
                });
                result.edits = mergeFileTextChanges(result.edits);
            }
            return result;
        },
        organizeImports(scope, formatOptions, preferences) {
            return choose(scope.fileName)
                .organizeImports(scope, formatOptions, preferences)
                .map((change) => {
                if (isVirtualFile(change.fileName))
                    change.fileName = getContainingFile(change.fileName);
                return change;
            });
        },
        getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences) {
            const suffix = '.vue' + VIRTUAL_FILENAME_SEPARATOR + '_module';
            const edits = choose(oldFilePath)
                .getEditsForFileRename(oldFilePath, newFilePath, formatOptions, preferences)
                .map((edit) => {
                if (isVirtualFile(edit.fileName)) {
                    const result = parseVirtualFileName(edit.fileName);
                    if (result == null)
                        return;
                    if (!['script', 'scriptSetup'].includes(result.selector.type))
                        return;
                    edit.fileName = getContainingFile(edit.fileName);
                }
                edit.textChanges = edit.textChanges
                    .map((change) => {
                    const index = change.newText.indexOf(suffix);
                    if (index >= 0) {
                        change.newText =
                            change.newText.substr(0, index + 4) +
                                change.newText.substr(index + suffix.length);
                    }
                    return change;
                })
                    .filter(isNotNull);
                if (edit.textChanges.length === 0)
                    return;
                return edit;
            })
                .filter(isNotNull);
            return mergeFileTextChanges(edits);
        },
        toggleLineComment(fileName, textRange) {
            return choose(fileName).toggleLineComment(fileName, textRange);
        },
        toggleMultilineComment(fileName, textRange) {
            return choose(fileName).toggleMultilineComment(fileName, textRange);
        },
        commentSelection(fileName, textRange) {
            return choose(fileName).commentSelection(fileName, textRange);
        },
        uncommentSelection(fileName, textRange) {
            return choose(fileName).uncommentSelection(fileName, textRange);
        },
        dispose() {
            config.context.disposeUnusedProjects(true);
            ts.dispose();
        },
    });
    return proxy;
}
function mergeFileTextChanges(edits) {
    const editsByFileName = {};
    edits.forEach((edit) => {
        if (!(edit.fileName in editsByFileName)) {
            editsByFileName[edit.fileName] = {
                fileName: edit.fileName,
                textChanges: [],
            };
        }
        const currentEdit = editsByFileName[edit.fileName];
        currentEdit.textChanges = combineTextChanges(currentEdit.textChanges, edit.textChanges);
        if (edit.isNewFile != null) {
            const value = currentEdit.isNewFile;
            currentEdit.isNewFile =
                value == null ? edit.isNewFile : value || edit.isNewFile;
        }
    });
    return Array.from(Object.values(editsByFileName));
}
function combineTextChanges(changes, newChanges) {
    const finalChanges = changes.slice();
    const ids = new Set(changes.map(serializeTextChange));
    newChanges.forEach((change) => {
        const id = serializeTextChange(change);
        if (!ids.has(id)) {
            ids.add(id);
            finalChanges.push(change);
        }
    });
    finalChanges.sort((a, b) => a.span.start - b.span.start);
    return finalChanges;
}
function serializeTextChange(change) {
    return JSON.stringify([change.span.start, change.span.length, change.newText]);
}

let context;
let server;
function init({ typescript }) {
    context = context !== null && context !== void 0 ? context : new PluginContext(typescript);
    server = server !== null && server !== void 0 ? server : new RoutingLanguageServer(context);
    return {
        create(info) {
            context.load(info);
            return server.decorate(info.languageService);
        },
        getExternalFiles(project) {
            var _a;
            return (_a = context.getExternalFiles(project)) !== null && _a !== void 0 ? _a : [];
        },
        onConfigurationChanged(config) {
            context.setConfig(config);
        },
    };
}

export default init;
//# sourceMappingURL=index.esm.js.map
